/* helpserv.c - Support Helper assistant service
 * Copyright 2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: helpserv.c,v 1.10 2002/05/25 02:28:20 sailorfrag Exp $
 */

/* TODO List for helpserv.c
 * - cmd_close should optionally take another argument (a reason/comment)
 * Wishlist for helpserv.c
 * - Let HelpServ accept "close nick|*handle", and make it work if there's
 *   only one request open to the target. complain and list request IDs if
 *   there are more than one.
 * - Make HelpServ send unassigned request count to helpers as they join the
 *   channel
 * - FAQ responses
 * - Get cmd_statsreport to sort by time
 * - A lot of the stuff (especially when talking to users) should be privmsg
 *   instead of notice, even if their handle is set otherwise
 * - Allow helpers to switch to help mode (I'm not entirely sure the use of
 *   this)
 * - A feature suggestion was to let it also show the waiting users in the
 *   help channel, in the order they joined, along with wait time (like
 *   eggdrop's .channel command)
 * - Something for users to use a subset of commands... like closing their
 *   own request(s), viewing what they've sent so far, and finding their
 *   helper's nick.
 * - It would be nice to have some variable expansions for the custom
 *   request open/close/etc messages. ${request/id}, etc
 * - Allow manipulation of the persist types on a per-request basis... so
 *   if it's normally set to close requests after a user quits, but there
 *   is a long-term issue, then that single request can remain
 * - Bot suspension
 */

#include "config.h"
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#include "dict.h"
#include "log.h"
#include "policer.h"
#include "helpfile.h"
#include "hash.h"
#include "nickserv.h"
#include "alias.h"
#include "opserv.h"
#include "parse.h"
#include "global.h"
#include "recdb.h"
#include "timeq.h"
#include "conf.h"
#include "helpserv.h"

#define HELPSERV_CONF_NAME "services/helpserv"

#define HELPSERV_DB_NAME "helpserv.db"
#define TEMP_DB_NAME "helpserv.db.new"
#define HELPSERV_HELPFILE_NAME "helpserv.help"

/* conf file options */
#define KEY_DESC "description"
#define KEY_DB_BACKUP_FREQ "db_backup_freq"

/* db entries */
#define KEY_BOTS "bots"
#define KEY_LAST_STATS_UPDATE "last_stats_update"
#define KEY_NICK "nick"
#define KEY_DB_BADCHANS "badchans"
#define KEY_HELP_CHANNEL "help_channel"
#define KEY_PAGE_DEST "page_dest"
#define KEY_CMDWORD "cmdword"
#define KEY_PERSIST_LENGTH "persist_length"
#define KEY_REGISTERED "registered"
#define KEY_REGISTRAR "registrar"
#define KEY_IDWRAP "id_wrap"
#define KEY_LAST_REQUESTID "last_requestid"
#define KEY_HELPERS "users"
#define KEY_HELPER_LEVEL "level"
#define KEY_HELPER_HELPMODE "helpmode"
#define KEY_HELPER_STATS "stats"
#define KEY_HELPER_STATS_TIME "time"
#define KEY_HELPER_STATS_PICKUP "picked_up"
#define KEY_HELPER_STATS_CLOSE "closed"
#define KEY_HELPER_STATS_REASSIGNFROM "reassign_from"
#define KEY_HELPER_STATS_REASSIGNTO "reassign_to"
#define KEY_REQUESTS "requests"
#define KEY_REQUEST_HELPER "helper"
#define KEY_REQUEST_HANDLE "handle"
#define KEY_REQUEST_TEXT "text"
#define KEY_REQUEST_OPENED "opened"
#define KEY_NOTIFICATION "notification"

/* General */
#define HSMSG_WRITE_SUCCESS	 "HelpServ db write completed (in "FMT_TIME_T".%03lu seconds)."
#define HSMSG_READHELP_SUCCESS	 "Read HelpServ help database in "FMT_TIME_T".%03ld seconds."
#define HSMSG_DISABLED           "HelpServ is disabled in the configuration file. If you enable it, you must restart srvx."
#define HSMSG_INVALID_BOT        "This command requires a valid HelpServ bot name."
#define HSMSG_ILLEGAL_CHANNEL    "$b%s$b is an illegal channel; cannot use it."
#define HSMSG_INTERNAL_COMMAND   "$b%s$b appears to be an internal HelpServ command, sorry."
#define HSMSG_NOT_IN_USERLIST    "%s lacks access to $b%s$b."
#define HSMSG_LEVEL_TOO_LOW      "You lack access to this command."
#define HSMSG_OPER_CMD           "This command can only be executed via $O."
#define HSMSG_NEED_CMD           "You must supply a HelpServ command. Try /msg $O helpserv help"
#define HSMSG_WTF_WHO_ARE_YOU    "$bUnable to find you on the %s userlist.$b This is a bug. Please report it."
#define HSMSG_NO_USE_OPSERV      "This command cannot be used via $O. If you really need to use it, add yourself to the userlist."
#define HSMSG_OPSERV_NEED_USER   "To use this command via $O, you must supply a user to target."
#define HSMSG_PAGE_REQUEST       "Page from $b%s$b: $b%s$b"
#define HSFMT_TIME               "%a, %d %b %Y %H:%M:%S %Z"
#define HSMSG_BAD_REQ_TYPE       "I don't know how to list $b%s$b requests."

/* Greetings */
#define HSMSG_GREET_EXISTING_REQ "Welcome back to %s. You have already opened request ID#%lu. Any messages you send to $S will be appended to that request."
#define HSMSG_GREET_PRIVMSG_EXISTREQ "Hello again. Your account has a previously opened request ID#%lu. This message and any further messages you send to $S will be appended to that request."

/* User Management */
#define HSMSG_CANNOT_ADD         "You do not have permission to add users."
#define HSMSG_CANNOT_DEL         "You do not have permission to delete users."
#define HSMSG_CANNOT_CLVL        "You do not have permission to modify users' access."
#define HSMSG_NO_SELF_CLVL       "You cannot change your own access."
#define HSMSG_NO_BUMP_ACCESS     "You cannot give users the same or more access than yourself."
#define HSMSG_NO_TRANSFER_SELF   "You cannot give ownership to your own accountv."
#define HSMSG_OWNERSHIP_GIVEN    "Ownership of $b%s$b has been transferred to account $b%s$b."
#define HSMSG_ADDED_USER         "Added new $b%s$b %s to the user list."
#define HSMSG_DELETED_USER       "Deleted %s $b%s$b %s from the user list."
#define HSMSG_INCORRECT_ACCESS   "%s is %s $b%s$b, not %s %s."
#define HSMSG_USER_EXISTS        "$b%s$b is already on the user list."
#define HSMSG_USER_IS_SERVICE    "$b%s$b is a network service and cannot be added to, or deleted from channel user lists."
#define HSMSG_INVALID_ACCESS     "$b%s$b is an invalid access level."
#define HSMSG_CHANGED_ACCESS     "%s is now %s $b%s$b"

/* Registration */
#define HSMSG_ALREADY_REGISTERED "The bot name %s has already been registered."
#define HSMSG_ILLEGAL_NICK       "$b%s$b is an illegal nick; cannot use it."
#define HSMSG_NICK_EXISTS        "The nick %s is in use by someone else."
#define HSMSG_REG_SUCCESS        "%s now has ownership of bot %s."
#define HSMSG_NEED_UNREG_CONFIRM "To unregister this bot, you must /msg $S unregister CONFIRM"

/* Rename */
#define HSMSG_RENAMED            "%s has been renamed to $b%s$b."
#define HSMSG_MOVED              "%s has been moved to $b%s$b."
#define HSMSG_INVALID_MOVE       "$b%s$b is not a valid nick or channel name."

/* Queue settings */
#define HSMSG_INVALID_OPTION     "$b%s$b is not a valid option."
#define HSMSG_QUEUE_OPTIONS      "HelpServ Queue Options:"
#define HSMSG_STRING_VALUE       "$b%19s $b%s"
#define HSMSG_PAGE_TARGET_VALUE  "$b%9s page dest $b%s"
#define HSMSG_PAGE_TYPE_VALUE    "$b%9s page type $b%s"
#define HSMSG_MESSAGE_VALUE      "$b%11s message $b%s"
#define HSMSG_INTERVAL_VALUE     "$b%19s $b%s"
#define HSMSG_PERSIST_VALUE      "$b%19s $b%s"
#define HSMSG_INVALID_INTERVAL   "Sorry, %s must be %s %s."
#define HSMSG_NEED_MANAGER       "Only managers or higher can do this."
#define HSMSG_SET_NEED_OPER      "This option can only be set by an oper."

/* Requests */
#define HSMSG_REQ_INVALID        "$b%s$b is not a valid request ID."
#define HSMSG_REQ_NOT_YOURS      "Request $b%lu$b is not assigned to you; it is currently %s%s."
#define HSMSG_REQ_CLOSED         "Request $b%lu$b has been closed."
#define HSMSG_REQ_NO_UNASSIGNED  "There are no unassigned requests."
#define HSMSG_REQ_NOTIN_HELPCHAN "%s cannot %s this request as %s are not in %s."
#define HSMSG_REQ_NOTIN_OVERRIDE "Note: %s are being assigned this request even though %s are not in %s, because the restriction was overridden (%s a manager or owner). If %s join and then part, this request will be marked unassigned."
#define HSMSG_REQ_ASSIGNED_YOU   "You have been assigned request ID#%lu:"
#define HSMSG_REQ_REASSIGNED     "You have been assigned request ID#%lu (reassigned from %s):"
#define HSMSG_REQ_INFO_1         "Request ID#%lu:"
#define HSMSG_REQ_INFO_2         " - Nick %s / Account %s"
#define HSMSG_REQ_INFO_3         " - Opened at %s (%s ago)"
#define HSMSG_REQ_INFO_4         " - Message:"
#define HSMSG_REQ_INFO_MESSAGE   "   %s"
#define HSMSG_REQ_ASSIGNED       "Your helper for request ID#%lu is %s (Current nick: %s)"
#define HSMSG_REQ_ASSIGNED_AGAIN "Your helper for request ID#%lu has been changed to %s (Current nick: %s)"
#define HSMSG_REQ_UNASSIGNED     "Your helper for request ID#%lu has %s, so your request is no longer being handled by them. It has been placed near the front of the unhandled request queue, based on how long ago your request was opened."
#define HSMSG_REQ_NEW            "Your help request (ID#%lu) has been recorded, and a helper should contact you shortly."
#define HSMSG_REQ_UNHANDLED_TIME "The first unhandled request has been waiting for %s."
#define HSMSG_REQ_NO_UNHANDLED   "There are no other unhandled requests."
#define HSMSG_REQ_PERSIST_QUIT   "Anything else you tell me until you are helped (or you quit) will be recorded as well.  If you disconnect, your request will be lost."
#define HSMSG_REQ_PERSIST_PART   "Anything else you tell me until you are helped (or you leave %s) will be recorded as well.  If you part from %s, your request will be lost."
#define HSMSG_REQ_PERSIST_HANDLE "Anything else you tell me until you are helped will be recorded as well."
#define HSMSG_REQ_FOUND_ANOTHER  "Request ID#%lu has been closed. $S detected that you also have request ID#%lu open. If you send $S a message, it will be associated with that request."

/* Messages that are inserted into request text */
#define HSMSG_REQMSG_STALE_UPDATE "[Stale request updated at %s]"
#define HSMSG_REQMSG_HELPER_NOTE  "[Helper note at %s]:"
#define HSMSG_REQMSG_HELPER_NOTE2 "  <%s> %s"
#define HSMSG_REQMSG_NOTE_ADDED   "Your note for request ID#%lu has been recorded."

/* Automatically generated page messages */
#define HSMSG_PAGE_NEW_REQUEST   "New request (ID#%lu) from $b%s$b (Account %s)"
#define HSMSG_PAGE_UPD_REQUEST   "Request ID#%lu has been updated by $b%s$b (Account %s). Request was initially opened at %s, and was last updated %s ago."
#define HSMSG_PAGE_CLOSE_REQUEST "Request ID#%lu from $b%s$b (Account %s) has been closed by %s."
#define HSMSG_PAGE_ASSIGN_REQUEST "Request ID#%lu from $b%s$b (Account %s) has been assigned to %s."
#define HSMSG_PAGE_HELPER_GONE   "Request ID#%lu from $b%s$b (Account %s) $bhas been unassigned$b, as its helper, %s has %s."
#if ANNOYING_ALERT_PAGES
#define HSMSG_PAGE_WHINE_HEADER  "$b%u unhandled requests$b waiting at least $b%s$b:"
#else
#define HSMSG_PAGE_WHINE_HEADER  "$b%u unhandled requests$b waiting at least $b%s$b."
#endif
#define HSMSG_PAGE_IDLE_HEADER   "$b%u users$b in %s $bidle at least %s$b:"
#define HSMSG_PAGE_EMPTYALERT    "$b%s has no helpers present$b"

/* Notification messages */
#define HSMSG_NOTIFY_USER_QUIT   "The user for request ID#%lu, $b%s$b, has disconnected."
#define HSMSG_NOTIFY_USER_MOVE   "The account for request ID#%lu, $b%s$b has been unregistered. It has been associated with user $b%s$b."
#define HSMSG_NOTIFY_USER_NICK   "The user for request ID#%lu has changed their nick from $b%s$b to $b%s$b."
#define HSMSG_NOTIFY_USER_FOUND  "The user for request ID#%lu is now online using nick $b%s$b."
#define HSMSG_NOTIFY_HAND_RENAME "The account for request ID#%lu has been renamed from $b%s$b to $b%s$b."
#define HSMSG_NOTIFY_HAND_MOVE   "The account for request ID#%lu has been changed to $b%s$b from $b%s$b."
#define HSMSG_NOTIFY_HAND_STUCK  "The user for request ID#%lu, $b%s$b, has re-authenticated to account $b%s$b from $b%s$b, and the request remained associated with the old handle."
#define HSMSG_NOTIFY_HAND_AUTH   "The user for request ID#%lu, $b%s$b, has authenticated to account $b%s$b."
#define HSMSG_NOTIFY_HAND_UNREG  "The account for request ID#%lu, $b%s$b, has been unregistered by $b%s$b."
#define HSMSG_NOTIFY_REQ_DROP    "Request ID#%lu has been $bdropped$b because %s."

/* Presence and request-related messages */
#define HSMSG_REQ_DROPPED_PART   "You have left $b%s$b. Your help request (ID#%lu) has been deleted."
#define HSMSG_REQ_WARN_UNREG     "The account you were authenticated to ($b%s$b) has been unregistered. Therefore unless you register a new handle, or authenticate to another one, if you disconnect, your HelpServ $S (%s) request ID#%lu will be lost."
#define HSMSG_REQ_DROPPED_UNREG  "By unregistering the account $b%s$b, HelpServ $S (%s) request ID#%lu was dropped, as there was nobody online to associate the request with."
#define HSMSG_REQ_ASSIGNED_UNREG "As the account $b%s$b was unregistered, HelpServ $S (%s) request ID#%lu was associated with you, as you were authenticated to that account. If you disconnect, then this request will be lost forever."
#define HSMSG_REQ_AUTH_STUCK     "By authenticating to account $b%s$b, HelpServ $S (%s) request ID#%lu remained with the previous account $b%s$b (because the request will remain until closed). This means that if you send $S a message, it $uwill not$u be associated with this request."
#define HSMSG_REQ_AUTH_MOVED     "By authenticating to account $b%s$b, HelpServ $S (%s) request ID#%lu ceased to be associated with your previously authenticated account ($b%s$b), and was transferred to your new account (because the request will only remain until you %s). This means that if you send $S a message, it $uwill$u be associated with this request."

/* Lists */
#define HSMSG_BOTLIST_HEADER     "$bCurrent HelpServ bots:$b"
#define HSMSG_USERLIST_HEADER    "$b%s users:$b"
#define HSMSG_USERLIST_NONE      "$bNone.$b"
#define HSMSG_USERLIST_ZOOT_LVL  "%s $b%ss$b:"
#define HSMSG_REQLIST_AUTH       "You are currently assigned these requests:"
#define HSMSG_REQ_LIST_TOP       "Listing $b%s$b requests."
#define HSMSG_REQ_LIST_NONE      "There are no matching requests."
#define HSMSG_STATS_TOP          "Stats for %s user $b%s$b:"
#define HSMSG_STATS_TIME         "$uTime spent helping in %s:$u"
#define HSMSG_STATS_REQS         "$uRequest activity statistics:$u"
#define HSFMT_STATS_REPORT_WEEK  "Stats report for the week of $b%a, %d %b %Y$b"

enum helpserv_level {
    HlNone,
    HlHelper,
    HlManager,
    HlOwner,
    HlOper,
    HlCount
};

static const char *helpserv_level_names[] = {
    "None",
    "Helper",
    "Manager",
    "Owner",
    "Oper",
    NULL
};

enum page_type {
    PAGE_NONE,
    PAGE_NOTICE,
    PAGE_PRIVMSG,
    PAGE_ONOTICE,
    PAGE_COUNT
};

static const char *page_type_names[] = {
    "None",
    "Notice",
    "Privmsg",
    "Onotice",
    NULL
};

static irc_send_func page_type_funcs[] = {
    NULL,
    irc_notice,
    irc_privmsg,
    irc_wallchops,
    NULL
};

enum page_source {
    PGSRC_COMMAND,
    PGSRC_ALERT,
    PGSRC_STATUS,
    PGSRC_COUNT
};

static const char *page_source_names[] = {
    "Command",
    "Alert",
    "Status",
    NULL
};

enum message_type {
    MSGTYPE_GREETING,
    MSGTYPE_REQ_OPENED,
    MSGTYPE_REQ_ASSIGNED,
    MSGTYPE_REQ_CLOSED,
    MSGTYPE_REQ_DROPPED,
    MSGTYPE_COUNT
};

static const char *message_type_names[] = {
    "Greeting",
    "ReqOpened",
    "ReqAssigned",
    "ReqClosed",
    "ReqDropped",
    NULL
};

enum interval_type {
    INTERVAL_IDLE_DELAY,
    INTERVAL_WHINE_DELAY,
    INTERVAL_WHINE_INTERVAL,
    INTERVAL_EMPTY_INTERVAL,
    INTERVAL_STALE_DELAY,
    INTERVAL_COUNT
};

static const char *interval_type_names[] = {
    "IdleDelay",
    "WhineDelay",
    "WhineInterval",
    "EmptyInterval",
    "StaleDelay",
    NULL
};

enum persistence_type {
    PERSIST_T_REQUEST,
    PERSIST_T_HELPER,
    PERSIST_T_COUNT
};

static const char *persistence_type_names[] = {
    "RequestPersist",
    "HelperPersist",
    NULL
};

enum persistence_length {
    PERSIST_PART,
    PERSIST_QUIT,
    PERSIST_CLOSE,
    PERSIST_COUNT
};

static const char *persistence_length_names[] = {
    "Part",
    "Quit",
    "Close",
    NULL
};

enum notification_type {
    NOTIFY_NONE,
    NOTIFY_DROP,
    NOTIFY_USER,
    NOTIFY_HANDLE,
    NOTIFY_COUNT
};

static const char *notification_type_names[] = {
    "None",
    "ReqDrop",
    "UserChanges",
    "AccountChanges",
    NULL
};

static struct {
    const char *description;
    unsigned long db_backup_frequency;
} helpserv_conf;

static int helpserv_enabled=0;
static time_t last_stats_update;

#define CMD_NEED_BOT            0x001
#define CMD_NOT_OVERRIDE        0x002
#define CMD_FROM_OPSERV_ONLY    0x004
#define CMD_IGNORE_EVENT        0x008
#define CMD_NEVER_FROM_OPSERV   0x010

struct helpserv_bot {
    struct userNode *helpserv;

    struct chanNode *helpchan;

    struct chanNode *page_targets[PGSRC_COUNT];
    enum page_type page_types[PGSRC_COUNT];
    char *messages[MSGTYPE_COUNT];
    unsigned long intervals[INTERVAL_COUNT];
    enum notification_type notify;

    /* This is a default; it can be changed on a per-request basis */
    enum persistence_type persist_types[PERSIST_T_COUNT];

    dict_t users; /* indexed by handle */

    struct helpserv_request *unhandled; /* linked list of unhandled requests */
    dict_t requests; /* indexed by request id */
    unsigned long last_requestid;
    unsigned long id_wrap;

    time_t registered;
    char *registrar;
};

struct helpserv_user {
    struct handle_info *handle;
    struct helpserv_bot *hs;
    unsigned int help_mode : 1;
    enum helpserv_level level;
    /* statistics */
    time_t join_time; /* when they joined, or 0 if not in channel */
    unsigned int time_per_week[4]; /* how long they've were in the channel the past 4 weeks */
    /* These statistics are kept both per-week (in [0]) and total (in [1]).  */
    unsigned int picked_up[2]; /* how many requests they have picked up */
    unsigned int closed[2]; /* how many requests they have closed */
    unsigned int reassigned_from[2]; /* how many requests reassigned from them to others */
    unsigned int reassigned_to[2]; /* how many requests reassigned from others to them */
};

struct helpserv_request {
    struct helpserv_user *helper;
    struct helpserv_bot *hs;
    struct string_list *text;
    struct helpserv_request *next_unhandled;

    struct helpserv_reqlist *parent_nick_list;
    struct helpserv_reqlist *parent_hand_list;

    /* One, but not both, of "user" and "handle" may be NULL,
     * depending on what we know about the user.
     *
     * If persist == PERSIST_CLOSE when the user quits, then it
     * switches to handle instead of user... and stays that way (it's
     * possible to have >1 nick per handle, so you can't really decide
     * to reassign a userNode to it unless they send another message
     * to HelpServ).
     */
    struct userNode *user;
    struct handle_info *handle;

    unsigned long id;
    time_t opened;
    time_t updated;
};

#define DEFINE_LIST_ALLOC(STRUCTNAME) \
struct STRUCTNAME * STRUCTNAME##_alloc() {\
    struct STRUCTNAME *newlist; \
    newlist = malloc(sizeof(struct STRUCTNAME)); \
    STRUCTNAME##_init(newlist); \
    return newlist; \
}\
void STRUCTNAME##_free(void *data) {\
    struct STRUCTNAME *list = data; /* void * to let dict_set_free_data use it */ \
    STRUCTNAME##_clean(list); \
    free(list); \
}

DECLARE_LIST(helpserv_botlist, struct helpserv_bot *);
DEFINE_LIST(helpserv_botlist, struct helpserv_bot *);
DEFINE_LIST_ALLOC(helpserv_botlist);

DECLARE_LIST(helpserv_reqlist, struct helpserv_request *);
DEFINE_LIST(helpserv_reqlist, struct helpserv_request *);
DEFINE_LIST_ALLOC(helpserv_reqlist);

DECLARE_LIST(helpserv_userlist, struct helpserv_user *);
DEFINE_LIST(helpserv_userlist, struct helpserv_user *);
DEFINE_LIST_ALLOC(helpserv_userlist);

const char *HSMSG_VERSION_ID = "$Id: helpserv.c,v 1.10 2002/05/25 02:28:20 sailorfrag Exp $";

struct helpfile *helpserv_helpfile;
static time_t lastwrite;
static dict_t helpserv_func_dict;
static dict_t helpserv_option_dict;
static dict_t helpserv_bots_dict; /* indexed by nick */
static dict_t helpserv_bots_bychan_dict; /* indexed by chan, holds a struct helpserv_botlist */
/* QUESTION: why are these outside of any helpserv_bot struct? */
static dict_t helpserv_reqs_bynick_dict; /* indexed by nick, holds a struct helpserv_reqlist */
static dict_t helpserv_reqs_byhand_dict; /* indexed by handle, holds a struct helpserv_reqlist */
static dict_t helpserv_users_byhand_dict; /* indexed by handle, holds a struct helpserv_userlist */

/* This is so that overrides can "speak" from opserv */
extern struct userNode *opserv;

#define HELPSERV_SYNTAX() 	helpserv_help(hs, from_opserv, user, argv[0], 0)
#define HELPSERV_FUNC(NAME) int NAME(struct userNode *user, struct helpserv_bot *hs, int from_opserv, unsigned int argc, unsigned char *argv[])
typedef HELPSERV_FUNC(helpserv_func_t);
#define HELPSERV_OPTION(NAME) HELPSERV_FUNC(NAME)
typedef HELPSERV_OPTION(helpserv_option_func_t);

static HELPSERV_FUNC(cmd_help);

#define REQUIRE_PARMS(N) if (argc < N) { \
        helpserv_notice(user, MSG_MISSING_PARAMS, argv[0]); \
        HELPSERV_SYNTAX(); \
        return 0; }

#define helpserv_notice(target, format...) send_message(target , (from_opserv ? opserv : hs->helpserv) , ## format)
#define helpserv_notify(helper, format...) do { struct userNode *_target; for (_target = (helper)->handle->users; _target; _target = _target->next_authed) { \
        send_message(_target, (helper)->hs->helpserv, ## format); \
    } } while (0)
#define helpserv_message(hs, target, id) do { if ((hs)->messages[id]) helpserv_notice(target, "%s", (hs)->messages[id]); } while (0)
#define helpserv_page(TYPE, FORMAT...) do { \
    struct chanNode *target=NULL; int msg_type=0; \
    target = hs->page_targets[TYPE]; \
    switch (hs->page_types[TYPE]) { \
        case PAGE_NOTICE: msg_type = 0; break; \
        case PAGE_PRIVMSG: msg_type = 1; break; \
        case PAGE_ONOTICE: msg_type = 2; break; \
        default: log(HS_LOG, LOG_ERROR, "helpserv_page() called but %s has an invalid page type %d\n", hs->helpserv->nick, TYPE); \
        case PAGE_NONE: target = NULL; break; \
    } \
    if (target) send_target_message(msg_type, target->name, hs->helpserv, ## FORMAT); \
    } while (0)
#define helpserv_get_handle_info(user, text) smart_get_handle_info((from_opserv ? opserv : hs->helpserv) , (user), (text))

struct helpserv_cmd {
    enum helpserv_level access;
    helpserv_func_t *func;
    double weight;
    long flags;
    unsigned int refs;
};

static const char * helpserv_level2str(enum helpserv_level level) {
    if (level <= HlOper) {
        return helpserv_level_names[level];
    } else {
        log(HS_LOG, LOG_ERROR, "helpserv_level2str receieved invalid level %d\n", level);
        return "Error";
    }
}

static enum helpserv_level helpserv_str2level(const char *msg) {
    enum helpserv_level nn;
    for (nn=HlNone; nn<=HlOper; nn++) {
        if (!strcasecmp(msg, helpserv_level_names[nn])) return nn;
    }
    log(HS_LOG, LOG_ERROR, "helpserv_str2level received invalid level %s\n", msg);
    return HlNone; /* Error */
}

static struct helpserv_user *GetHSUser(struct helpserv_bot *hs, struct handle_info *hi) {
    return dict_find(hs->users, hi->handle, NULL);
}

/* Handle a message from a user to a HelpServ bot. */
static void helpserv_usermsg(struct userNode *user, struct helpserv_bot *hs, char *text) {
    const int from_opserv = 0; /* for helpserv_notice */
    struct helpserv_request *req=NULL, *newest=NULL;
    struct helpserv_reqlist *reqlist, *hand_reqlist;
    unsigned int i;

    if ((reqlist = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL))) {
        for (i=0; i < reqlist->used; i++) {
            req = reqlist->list[i];
            if (req->hs != hs) {
                continue;
            }
            if (!newest || (newest->opened < req->opened)) {
                newest = req;
            }
        }

        /* If nothing was found, this will set req to NULL */
        req = newest;
    }

    if (user->handle_info) {
        hand_reqlist = dict_find(helpserv_reqs_byhand_dict, user->handle_info->handle, NULL);
        if (!req && hand_reqlist) {
            /* Most recent first again */
            for (i=0; i < hand_reqlist->used; i++) {
                req = hand_reqlist->list[i];
                if (req->hs != hs) {
                    continue;
                }
                if (!newest || (newest->opened < req->opened)) {
                    newest = req;
                }
            }
            req = newest;

            if (req) {
                req->user = user;
                if (!reqlist) {
                    reqlist = helpserv_reqlist_alloc();
                    dict_insert(helpserv_reqs_bynick_dict, user->nick, reqlist);
                }
                req->parent_nick_list = reqlist;
                helpserv_reqlist_append(reqlist, req);

                if (req->helper && (hs->notify >= NOTIFY_USER)) helpserv_notify(req->helper, HSMSG_NOTIFY_USER_FOUND, req->id, user->nick);

                helpserv_notice(user, HSMSG_GREET_PRIVMSG_EXISTREQ, req->id);
            }
        }
    } else {
        hand_reqlist = NULL;
    }

    if (!req) {
        char lbuf[3][MAX_LINE_SIZE], unh[INTERVALLEN];

        if ((hs->persist_types[PERSIST_T_REQUEST] == PERSIST_PART) && !GetUserMode(hs->helpchan, user)) {
            helpserv_notice(user, HSMSG_REQ_NOTIN_HELPCHAN, "You", "open", "you", hs->helpchan->name);
            return;
        }

        req = calloc(1, sizeof(struct helpserv_request));

        req->id = ++hs->last_requestid;
        sprintf(unh, "%lu", req->id);
        dict_insert(hs->requests, strdup(unh), req);

        if (hs->id_wrap) {
            if (hs->last_requestid >= hs->id_wrap) {
                unsigned long i;
                for (i=1; i < hs->id_wrap; i++) {
                    char buf[12];
                    sprintf(buf, "%lu", i);
                    if (!dict_find(hs->requests, buf, NULL)) {
                        hs->last_requestid = i-1;
                        break;
                    }
                }
                if (i >= hs->id_wrap) {
                    log(HS_LOG, LOG_INFO, "%s has more requests than its id_wrap\n", hs->helpserv->nick);
                }
            }
        }

        req->hs = hs;
        req->helper = NULL;
        req->text = alloc_string_list(8);
        req->user = user;
        req->handle = user->handle_info;
        req->opened = now;

        if (!hs->unhandled) {
            hs->unhandled = req;
            req->next_unhandled = NULL;
        } else {
            /* Set req->next_unhandled temporarily to be the previous
             * unhandled request, then set it null (last in list) */
            for (req->next_unhandled = hs->unhandled; req->next_unhandled->next_unhandled; req->next_unhandled = req->next_unhandled->next_unhandled);
            req->next_unhandled->next_unhandled = req;
            req->next_unhandled = NULL;
        }

        if (!reqlist) {
            reqlist = helpserv_reqlist_alloc();
            dict_insert(helpserv_reqs_bynick_dict, user->nick, reqlist);
        }
        req->parent_nick_list = reqlist;
        helpserv_reqlist_append(reqlist, req);

        if (user->handle_info) {
            if (!hand_reqlist) {
                hand_reqlist = helpserv_reqlist_alloc();
                dict_insert(helpserv_reqs_byhand_dict, user->handle_info->handle, hand_reqlist);
            }
            req->parent_hand_list = hand_reqlist;
            helpserv_reqlist_append(hand_reqlist, req);
        } else {
            req->parent_hand_list = NULL;
        }

        sprintf(lbuf[0], HSMSG_REQ_NEW, req->id);
        if (req->next_unhandled) {
            intervalString(unh, now - req->next_unhandled->opened);
            sprintf(lbuf[1], HSMSG_REQ_UNHANDLED_TIME, unh);
        } else {
            sprintf(lbuf[1], HSMSG_REQ_NO_UNHANDLED);
        }
        switch (hs->persist_types[PERSIST_T_REQUEST]) {
            case PERSIST_PART:
                sprintf(lbuf[2], HSMSG_REQ_PERSIST_PART, hs->helpchan->name, hs->helpchan->name);
                break;
            case PERSIST_QUIT:
                sprintf(lbuf[2], HSMSG_REQ_PERSIST_QUIT);
                break;
            default:
                log(HS_LOG, LOG_ERROR, "%s has an invalid req_persist\n", hs->helpserv->nick);
            case PERSIST_CLOSE:
                if (user->handle_info) {
                    sprintf(lbuf[2], HSMSG_REQ_PERSIST_HANDLE);
                } else {
                    sprintf(lbuf[2], HSMSG_REQ_PERSIST_QUIT);
                }
                break;
        }
        helpserv_message(hs, user, MSGTYPE_REQ_OPENED);
        helpserv_notice(user, "%s %s %s", lbuf[0], lbuf[1], lbuf[2]);
        helpserv_page(PGSRC_STATUS, HSMSG_PAGE_NEW_REQUEST, req->id, user->nick, user->handle_info ? user->handle_info->handle : "not authed");
    } else if (hs->intervals[INTERVAL_STALE_DELAY] && (req->updated < TIME_T_CAST(now - hs->intervals[INTERVAL_STALE_DELAY]))) {
        char buf[MAX_LINE_SIZE], updatestr[INTERVALLEN], timestr[MAX_LINE_SIZE];
        strftime(timestr, MAX_LINE_SIZE, HSFMT_TIME, localtime(&req->opened));
        intervalString(updatestr, now - req->updated);
        if (req->helper && (hs->notify >= NOTIFY_USER)) {
            helpserv_notify(req->helper, HSMSG_PAGE_UPD_REQUEST, req->id, user->nick, user->handle_info ? user->handle_info->handle : "not authed", timestr, updatestr);
        } else {
            helpserv_page(PGSRC_STATUS, HSMSG_PAGE_UPD_REQUEST, req->id, user->nick, user->handle_info ? user->handle_info->handle : "not authed", timestr, updatestr);
        }
        strftime(timestr, MAX_LINE_SIZE, HSFMT_TIME, localtime(&now));
        snprintf(buf, MAX_LINE_SIZE, HSMSG_REQMSG_STALE_UPDATE, timestr);
        string_list_append(req->text, strdup(buf));
    }

    req->updated = now;
    string_list_append(req->text, strdup(text));
}

/* Handle messages direct to a HelpServ bot. */
static void helpserv_botmsg(struct userNode *user, struct userNode *target, char *text, int server_qualified) {
    struct helpserv_bot *hs;
    struct helpserv_cmd *cmd;
    struct helpserv_user *hs_user;
    unsigned char *argv[MAXNUMPARAMS];
    int argc, argv_shift;
    const int from_opserv = 0; /* for helpserv_notice */

    (void)server_qualified;

    /* Ignore things consisting of empty lines */
    if (!*text) return;

    hs = dict_find(helpserv_bots_dict, target->nick, NULL);

    if (!policer_conforms(user->command_policer, now, 1)) {
        helpserv_notice(user, MSG_COMMAND_FLOOD);
        return;
    }

    /* See if we should listen to their message as a command (helper)
     * or a help request (user) */
    if (!user->handle_info || !(hs_user = dict_find(hs->users, user->handle_info->handle, NULL))) {
        helpserv_usermsg(user, hs, text);
        return;
    }

    argv_shift = 1;
    argc = split_line(text, false, ArrayLength(argv)-argv_shift, argv+argv_shift);
    if (!argc) return;

    cmd = dict_find(helpserv_func_dict, argv[argv_shift], NULL);
    if (!cmd) {
        helpserv_notice(user, MSG_COMMAND_UNKNOWN, argv[argv_shift]);
        return;
    }
    if (cmd->flags & CMD_FROM_OPSERV_ONLY) {
        helpserv_notice(user, HSMSG_OPER_CMD);
        return;
    }
    if (cmd->access > hs_user->level) {
        helpserv_notice(user, HSMSG_LEVEL_TOO_LOW);
        return;
    }
    if (cmd->func) {
        char botnick[NICKLEN+1];
        strcpy(botnick, hs->helpserv->nick);

        if (cmd->func(user, hs, 0, argc, argv+argv_shift)) {
            log(HS_LOG, LOG_INFO, "%s[%s] (%s): %s\n", user->nick, user->handle_info->handle, botnick, text);
        }
    } else {
        helpserv_notice(user, HSMSG_INTERNAL_COMMAND, argv[argv_shift]);
    }
}

/* Handle a control command from an IRC operator */
int helpserv_oper_cmd(struct userNode *user, int argc, unsigned char *argv[]) {
    struct helpserv_bot *hs = NULL;
    struct helpserv_cmd *cmd;
    const int from_opserv = 1; /* for helpserv_notice */
    unsigned char *cmd_argv[MAXNUMPARAMS];
    char botnick[NICKLEN+1];
    int retval, i;

    if (!helpserv_enabled) {
        helpserv_notice(user, HSMSG_DISABLED);
        return 0;
    }

    if (argc < 1) {
        send_help(user, opserv, helpserv_helpfile, NULL);
        return 0;
    }

    if (!(cmd = dict_find(helpserv_func_dict, argv[0], NULL))) {
        helpserv_notice(user, MSG_COMMAND_UNKNOWN, argv[0]);
        return 0;
    }

    if ((cmd->flags & CMD_NEED_BOT) && ((argc < 2) || !(hs = dict_find(helpserv_bots_dict, argv[1], NULL)))) {
        helpserv_notice(user, HSMSG_INVALID_BOT);
        return 0;
    }

    if (cmd->flags & CMD_NEVER_FROM_OPSERV) {
        helpserv_notice(user, HSMSG_NO_USE_OPSERV);
        return 0;
    }

    if (cmd->func) {
        if (hs) {
            cmd_argv[0] = argv[0];
            for (i=2; i<argc; i++) {
                cmd_argv[i-1] = argv[i];
            }
            strcpy(botnick, hs->helpserv->nick);
            retval = cmd->func(user, hs, 1, argc-1, cmd_argv);
        } else {
            strcpy(botnick, "No Bot");
            retval = cmd->func(user, hs, 1, argc, argv);
        }
    } else {
        helpserv_notice(user, HSMSG_INTERNAL_COMMAND, argv[0]);
        return 0;
    }

    if (retval && !(cmd->flags & CMD_IGNORE_EVENT)) {
        unsplit_string(argv, argc);
        log(HS_LOG, LOG_INFO, "%s[%s] (%s): %s%s\n", user->nick, user->handle_info->handle, botnick, argv[0], (cmd->flags & (CMD_NOT_OVERRIDE|CMD_FROM_OPSERV_ONLY) ? "" : " (override)"));
    }

    return retval;
}

static HELPSERV_FUNC(cmd_version) {
    (void)hs;

    helpserv_notice(user, MSG_VERSION);

    if ((argc > 1 && !strcasecmp(argv[1], "cvs")) || from_opserv) {
        helpserv_notice(user, "%s", HSMSG_VERSION_ID);
    }

    return 1;
}

static void helpserv_help(struct helpserv_bot *hs, int from_opserv, struct userNode *user, const char *topic, int show_aliases) {
    char *aliases, *resolved = resolve_alias("HelpServ", (char *)topic);

    send_help(user, (from_opserv ? opserv : hs->helpserv), helpserv_helpfile, resolved);
    if (show_aliases && (aliases = command_aliases("HelpServ", resolved))) {
        char *c = aliases;

        for (c = aliases; *c; c++) {
            *c = toupper(*c);
        }

        helpserv_notice(user, "Aliases: %s", aliases);
        free(aliases);
    }
}

static int append_entry(const char *key, void *data, void *extra) {
    struct helpfile_expansion *exp = extra;
    int row;

    (void)data;
    row = exp->value.table.length++;
    exp->value.table.contents[row] = calloc(1, sizeof(char*));
    exp->value.table.contents[row][0] = key;
    return 0;
}

static struct helpfile_expansion helpserv_expand_variable(char *variable) {
    struct helpfile_expansion exp;

    if (!strcasecmp(variable, "index")) {
        exp.type = HF_TABLE;
        exp.value.table.length = 1;
        exp.value.table.width = 1;
        exp.value.table.flags = TABLE_REPEAT_ROWS;
        exp.value.table.contents = calloc(dict_size(helpserv_func_dict)+1, sizeof(char**));
        exp.value.table.contents[0] = calloc(1, sizeof(char*));
        exp.value.table.contents[0][0] = "Commands:";
        dict_foreach(helpserv_func_dict, append_entry, &exp);
        return exp;
    }

    exp.type = HF_STRING;
    exp.value.str = NULL;
    return exp;
}

static void helpserv_helpfile_read(void) {
    helpserv_helpfile = open_helpfile(HELPSERV_HELPFILE_NAME, helpserv_expand_variable);
}

static HELPSERV_FUNC(cmd_help) {
    const char *topic;

    if (argc < 2) {
        topic = NULL;
    } else {
        topic = argv[1];
        unsplit_string(argv+1, argc-1);
    }
    helpserv_help(hs, from_opserv, user, topic, 1);

    return 1;
}

static HELPSERV_FUNC(cmd_readhelp) {
    struct timeval start, stop;
    struct helpfile *old_helpfile = helpserv_helpfile;

    (void)argc; (void)argv; (void)hs; /* ignore them! ignore them all! */

    gettimeofday(&start, NULL);
    helpserv_helpfile_read();
    if (helpserv_helpfile) {
        close_helpfile(old_helpfile);
    } else {
        helpserv_helpfile = old_helpfile;
    }
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
        stop.tv_sec -= 1;
        stop.tv_usec += 1000000;
    }
    helpserv_notice(user, HSMSG_READHELP_SUCCESS, stop.tv_sec, stop.tv_usec/1000);

    return 1;
}

static struct helpserv_user * helpserv_add_user(struct helpserv_bot *hs, struct handle_info *handle, enum helpserv_level level) {
    struct helpserv_user *hs_user;
    struct helpserv_userlist *userlist;

    hs_user = calloc(1, sizeof(struct helpserv_user));
    hs_user->handle = handle;
    hs_user->hs = hs;
    hs_user->help_mode = 0;
    hs_user->level = level;
    hs_user->join_time = find_handle_in_channel(hs->helpchan, handle, NULL) ? now : 0;
    dict_insert(hs->users, handle->handle, hs_user);

    if (!(userlist = dict_find(helpserv_users_byhand_dict, handle->handle, NULL))) {
        userlist = helpserv_userlist_alloc();
        dict_insert(helpserv_users_byhand_dict, handle->handle, userlist);
    }
    helpserv_userlist_append(userlist, hs_user);

    return hs_user;
}

static void helpserv_del_user(struct helpserv_bot *hs, struct helpserv_user *hs_user) {
    dict_remove(hs->users, hs_user->handle->handle);
}

static int cmd_add_user(struct helpserv_bot *hs, int from_opserv, struct userNode *user, enum helpserv_level level, int argc, unsigned char *argv[]) {
    struct helpserv_user *actor, *new_user;
    struct handle_info *handle;

    REQUIRE_PARMS(2);

    if (!from_opserv) {
        actor = GetHSUser(hs, user->handle_info);
        if (actor->level < HlManager) {
            helpserv_notice(user, HSMSG_CANNOT_ADD);
            return 0;
        }
    } else {
        actor = NULL;
    }

    if (!(handle = helpserv_get_handle_info(user, argv[1]))) return 0;

    if (GetHSUser(hs, handle)) {
        helpserv_notice(user, HSMSG_USER_EXISTS, handle->handle);
        return 0;
    }

    if (!(from_opserv) && actor && (actor->level <= level)) {
        helpserv_notice(user, HSMSG_NO_BUMP_ACCESS);
        return 0;
    }

    new_user = helpserv_add_user(hs, handle, level);

    helpserv_notice(user, HSMSG_ADDED_USER, helpserv_level2str(level), handle->handle);
    return 1;
}

static HELPSERV_FUNC(cmd_deluser) {
    struct helpserv_user *actor=NULL, *victim;
    struct handle_info *handle;
    enum helpserv_level level;

    REQUIRE_PARMS(2);

    if (!from_opserv) {
        actor = GetHSUser(hs, user->handle_info);
        if (actor->level < HlManager) {
            helpserv_notice(user, HSMSG_CANNOT_DEL);
            return 0;
        }
    }

    if (!(handle = helpserv_get_handle_info(user, argv[1]))) return 0;

    if (!(victim = GetHSUser(hs, handle))) {
        helpserv_notice(user, HSMSG_NOT_IN_USERLIST, handle->handle, hs->helpserv->nick);
        return 0;
    }

    if (!from_opserv && actor && (actor->level <= victim->level)) {
        helpserv_notice(user, MSG_USER_OUTRANKED, victim->handle->handle);
        return 0;
    }

    level = victim->level;
    helpserv_del_user(hs, victim);
    helpserv_notice(user, HSMSG_DELETED_USER, preposition(helpserv_level2str(level)), helpserv_level2str(level), handle->handle);
    return 1;
}

static HELPSERV_FUNC(cmd_helpers) {
    struct helpserv_user **userList[HlCount];
    enum helpserv_level lvl;
    struct helpfile_table tbl;
    dict_iterator_t it;
    unsigned int line, i;
    unsigned int num_users[HlCount+1]={ 0 }; /* num_users[HlCount] == total
                                                for all levels between
                                                lvl_lowest and lvl_highest */

    /* If we want to add stuff like "wlist", "mlist", and "hlist", make it a
     * copy+paste type job */
    const enum helpserv_level lvl_lowest=HlHelper, lvl_highest=HlOwner;

    (void)argc; (void)argv;

    helpserv_notice(user, HSMSG_USERLIST_HEADER, hs->helpserv->nick);
    if (dict_size(hs->users) == 0) {
        helpserv_notice(user, HSMSG_USERLIST_NONE);
        return 1;
    }

    for (it=dict_first(hs->users); it; it=iter_next(it)) {
        struct helpserv_user *hs_user=iter_data(it);
        num_users[hs_user->level]++;
    }

    for (lvl = HlNone; lvl < HlCount; lvl++) {
        if (num_users[lvl] && (lvl >= lvl_lowest) && (lvl <= lvl_highest)) {
            userList[lvl] = alloca(num_users[lvl] * sizeof(**userList));
            num_users[HlCount] += num_users[lvl];
        } else {
            userList[lvl] = NULL;
        }
        /* Reset for next use; won't reset num_users[HlCount] */
        num_users[lvl] = 0;
    }

    /* If there's users with level < lvl_lowest, > lvl_highest, and none in
     * between */
    if (!num_users[HlCount]) {
        helpserv_notice(user, HSMSG_USERLIST_NONE);
        return 1;
    }

    for (it=dict_first(hs->users); it; it=iter_next(it)) {
        struct helpserv_user *hs_user=iter_data(it);
        
        if ((hs_user->level < lvl_lowest) || (hs_user->level > lvl_highest)) {
            continue;
        }

        userList[hs_user->level][num_users[hs_user->level]++] = hs_user;
    }

    switch (user->handle_info->userlist_style) {
        case HI_STYLE_DEF:
            line = 1;
            tbl.length = dict_size(hs->users)+1;
            tbl.width = 2;
            tbl.flags = TABLE_NO_FREE;
            tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
            tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
            tbl.contents[0][0] = "Level";
            tbl.contents[0][1] = "Handle";

            for (lvl = lvl_highest; lvl >= lvl_lowest; lvl--) {
                for (i=0; i < num_users[lvl]; i++, line++) {
                    tbl.contents[line] = alloca(tbl.width * sizeof(**tbl.contents));
                    tbl.contents[line][0] = helpserv_level_names[lvl];
                    tbl.contents[line][1] = userList[lvl][i]->handle->handle;
                }
            }

            table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);

            break;
        case HI_STYLE_ZOOT:
        default:
            /* Technically this isn't true Zoot style (it would only have one
             * user per line), but there wasn't anything else useful to show
             * besides the users' handle names. */
            for (lvl = lvl_highest; lvl >= lvl_lowest; lvl--) {
                if (!num_users[lvl]) continue;

                helpserv_notice(user, HSMSG_USERLIST_ZOOT_LVL, hs->helpserv->nick, helpserv_level_names[lvl]);

                tbl.length = num_users[lvl];
                tbl.width = 1;
                tbl.flags = TABLE_NO_FREE | TABLE_REPEAT_ROWS | TABLE_NO_HEADERS;
                tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));

                for (i=0; i < num_users[lvl]; i++) {
                    tbl.contents[i] = alloca(tbl.width * sizeof(**tbl.contents));
                    tbl.contents[i][0] = userList[lvl][i]->handle->handle;
                }

                table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);
            }
    }
    return 1;
}

static HELPSERV_FUNC(cmd_addowner) {
    return cmd_add_user(hs, from_opserv, user, HlOwner, argc, argv);
}

static HELPSERV_FUNC(cmd_addmanager) {
    return cmd_add_user(hs, from_opserv, user, HlManager, argc, argv);
}

static HELPSERV_FUNC(cmd_addhelper) {
    return cmd_add_user(hs, from_opserv, user, HlHelper, argc, argv);
}

static HELPSERV_FUNC(cmd_clvl) {
    struct helpserv_user *actor=NULL, *victim;
    struct handle_info *handle;
    enum helpserv_level level;

    REQUIRE_PARMS(3);

    if (!from_opserv) {
        actor = GetHSUser(hs, user->handle_info);
        if (actor->level < HlOwner) {
            helpserv_notice(user, HSMSG_CANNOT_CLVL);
            return 0;
        }
    }

    if (!(handle = helpserv_get_handle_info(user, argv[1]))) return 0;

    if (!(victim = GetHSUser(hs, handle))) {
        helpserv_notice(user, HSMSG_NOT_IN_USERLIST, handle->handle, hs->helpserv->nick);
        return 0;
    }

    if (((level = helpserv_str2level(argv[2])) == HlNone) || level == HlOper) {
        helpserv_notice(user, HSMSG_INVALID_ACCESS, argv[2]);
        return 0;
    }

    if (!(from_opserv) && actor) {
        if (actor == victim) {
            helpserv_notice(user, HSMSG_NO_SELF_CLVL);
            return 0;
        }

        if (actor->level <= victim->level) {
            helpserv_notice(user, MSG_USER_OUTRANKED, victim->handle->handle);
            return 0;
        }

        if (level >= actor->level) {
            helpserv_notice(user, HSMSG_NO_BUMP_ACCESS);
            return 0;
        }
    }

    victim->level = level;
    helpserv_notice(user, HSMSG_CHANGED_ACCESS, handle->handle, preposition(helpserv_level2str(level)), helpserv_level2str(level));

    return 1;
}

static void free_request(void *data) {
    struct helpserv_request *req = data;

    /* Clean up from the unhandled queue */
    if (req->hs->unhandled) {
        if (req->hs->unhandled == req) {
            req->hs->unhandled = req->next_unhandled;
        } else {
            struct helpserv_request *uh;
            for (uh=req->hs->unhandled; uh->next_unhandled && (uh->next_unhandled != req); uh = uh->next_unhandled);
            if (uh->next_unhandled) {
                uh->next_unhandled = req->next_unhandled;
            }
        }
    }

    /* Clean up the lists */
    if (req->parent_nick_list) {
        if (req->parent_nick_list->used == 1) {
            dict_remove(helpserv_reqs_bynick_dict, req->user->nick);
        } else {
            helpserv_reqlist_remove(req->parent_nick_list, req);
        }
    }
    if (req->parent_hand_list) {
        if (req->parent_hand_list->used == 1) {
            dict_remove(helpserv_reqs_byhand_dict, req->handle->handle);
        } else {
            helpserv_reqlist_remove(req->parent_hand_list, req);
        }
    }

    free_string_list(req->text);
    free(req);
}

static HELPSERV_FUNC(cmd_close) {
    struct helpserv_request *req, *newest=NULL;
    struct helpserv_reqlist *nick_list, *hand_list;
    unsigned long old_req;
    unsigned int i;

    /* TODO: cmd_close should optionally take another argument (a reason/comment) */
    REQUIRE_PARMS(2);

    if (!(req = dict_find(hs->requests, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_REQ_INVALID, argv[1]);
        return 0;
    }

    if (!req->helper || (req->helper->handle != user->handle_info)) {
        helpserv_notice(user, HSMSG_REQ_NOT_YOURS, req->id, (req->helper ? "assigned to " : "unassigned"), (req->helper ? req->helper->handle->handle : ""));
        return 0;
    }

    if (req->user) helpserv_message(hs, req->user, MSGTYPE_REQ_CLOSED);
    helpserv_notice(user, HSMSG_REQ_CLOSED, req->id);
    helpserv_page(PGSRC_STATUS, HSMSG_PAGE_CLOSE_REQUEST, req->id, req->user ? req->user->nick : "an offline user", req->handle ? req->handle->handle : "not authed", user->nick);

    req->helper->closed[0]++;
    req->helper->closed[1]++;

    /* Set these to keep track of the lists after the request is gone, but
     * not if free_request() will helpserv_reqlist_free() them. */
    nick_list = req->parent_nick_list;
    if (nick_list && (nick_list->used == 1)) nick_list = NULL;
    hand_list = req->parent_hand_list;
    if (hand_list && (hand_list->used == 1)) hand_list = NULL;
    old_req = req->id;

    dict_remove(hs->requests, argv[1]);

    /* Look for other requests associated with them */
    if (nick_list) {
        for (i=0; i < nick_list->used; i++) {
            req = nick_list->list[i];

            if (req->hs != hs) {
                continue;
            }
            if (!newest || (newest->opened < req->opened)) {
                newest = req;
            }
        }

        if (newest) {
            helpserv_notice(newest->user, HSMSG_REQ_FOUND_ANOTHER, old_req, newest->id);
        }
    }

    return 1;
}

static HELPSERV_FUNC(cmd_list) {
    dict_iterator_t it;
    int searchtype;
    struct helpfile_table tbl;
    unsigned int line, total;
    struct helpserv_request *req;

    if ((argc < 2) || !strcasecmp(argv[1], "unassigned")) {
        helpserv_notice(user, HSMSG_REQ_LIST_TOP, "all unassigned");
        for (req = hs->unhandled, total=0; req; req = req->next_unhandled, total++);
        searchtype = 1; /* Unassigned */
    } else if (!strcasecmp(argv[1], "assigned")) {
        helpserv_notice(user, HSMSG_REQ_LIST_TOP, "all assigned");
        for (req = hs->unhandled, total=dict_size(hs->requests); req; req = req->next_unhandled, total--);
        searchtype = 2; /* Assigned */
    } else if (!strcasecmp(argv[1], "me")) {
        helpserv_notice(user, HSMSG_REQ_LIST_TOP, "your");
        for (total = 0, it = dict_first(hs->requests); it; it = iter_next(it)) {
            req = iter_data(it);
            if (req->helper && (req->helper->handle == user->handle_info)) total++;
        }
        searchtype = 4;
    } else if (!strcasecmp(argv[1], "all")) {
        helpserv_notice(user, HSMSG_REQ_LIST_TOP, "all");
        total = dict_size(hs->requests);
        searchtype = 3; /* All */
    } else {
        helpserv_notice(user, HSMSG_BAD_REQ_TYPE, argv[1]);
        return 0;
    }

    if (!total) {
        helpserv_notice(user, HSMSG_REQ_LIST_NONE);
        return 1;
    }

    tbl.length = total+1;
    tbl.width = 5;
    tbl.flags = TABLE_NO_FREE;
    tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
    tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[0][0] = "ID#";
    tbl.contents[0][1] = "User";
    tbl.contents[0][2] = "Helper";
    tbl.contents[0][3] = "Time open";
    tbl.contents[0][4] = "User status";

    for (it=dict_first(hs->requests), line=0; it; it=iter_next(it)) {
        char opentime[INTERVALLEN], reqid[12], username[NICKLEN+1];

        req = iter_data(it);

        switch (searchtype) {
            case 1:
                if (req->helper) continue;
                break;
            case 2:
                if (!req->helper) continue;
                break;
            case 3:
            default:
                break;
            case 4:
                if (req->helper->handle != user->handle_info) continue;
                break;
        }

        line++;

        tbl.contents[line] = alloca(tbl.width * sizeof(**tbl.contents));
        sprintf(reqid, "%lu", req->id);
        tbl.contents[line][0] = strdup(reqid);
        if (req->user) {
            strcpy(username, req->user->nick);
        } else {
            username[0] = '*';
            strcpy(username+1, req->handle->handle);
        }
        tbl.contents[line][1] = strdup(username);
        tbl.contents[line][2] = req->helper ? req->helper->handle->handle : "(Unassigned)";
        intervalString(opentime, now - req->opened);
        tbl.contents[line][3] = strdup(opentime);
        tbl.contents[line][4] = ((req->user || req->handle->users) ? "Online" : "Offline");
    }

    table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);

    for (; line > 0; line--) {
        free((char *)tbl.contents[line][0]);
        free((char *)tbl.contents[line][1]);
        free((char *)tbl.contents[line][3]);
    }

    return 1;
}

static void helpserv_show(int from_opserv, struct helpserv_bot *hs, struct userNode *user, struct helpserv_request *req) {
    unsigned int nn;
    char buf[MAX_LINE_SIZE];
    char buf2[INTERVALLEN];

    helpserv_notice(user, HSMSG_REQ_INFO_2, (req->user ? req->user->nick : ((req->handle && req->handle->users) ? "unknown, but online" : "not online")), (req->handle ? req->handle->handle : "not authed"));
    strftime(buf, MAX_LINE_SIZE, HSFMT_TIME, localtime(&req->opened));
    intervalString(buf2, now - req->opened);
    helpserv_notice(user, HSMSG_REQ_INFO_3, buf, buf2);
    helpserv_notice(user, HSMSG_REQ_INFO_4);
    for (nn=0; nn < string_list_used(req->text); nn++) {
        helpserv_notice(user, HSMSG_REQ_INFO_MESSAGE, string_list_get(req->text, nn));
    }
}

/* actor is the one who executed the command... it should == user except from
 * cmd_assign */
static int helpserv_assign(int from_opserv, struct helpserv_bot *hs, struct userNode *user, struct userNode *actor, struct helpserv_request *req) {
    struct helpserv_request *req2;
    struct helpserv_user *old_helper;

    if (!user->handle_info) return 0;
    if ((hs->persist_types[PERSIST_T_HELPER] == PERSIST_PART) && !GetUserMode(hs->helpchan, user)) {
        struct helpserv_user *hsuser_actor = GetHSUser(hs, actor->handle_info);
        if (hsuser_actor->level >= HlManager) {
            helpserv_notice(user, HSMSG_REQ_NOTIN_OVERRIDE, "You", "you", hs->helpchan->name, (actor == user ? "you are" : "by"), "you");
            if (user != actor) {
                helpserv_notice(actor, HSMSG_REQ_NOTIN_OVERRIDE, "They", "they", hs->helpchan->name, "as you are", "they");
            }
        } else {
            helpserv_notice(user, HSMSG_REQ_NOTIN_HELPCHAN, "You", "be assigned", "you", hs->helpchan->name);
            return 0;
        }
    }

    if ((old_helper = req->helper)) {
        /* don't need to remove from the unhandled queue */
    } else if (hs->unhandled == req) {
        hs->unhandled = req->next_unhandled;
    } else for (req2 = hs->unhandled; req2; req2 = req2->next_unhandled) {
        if (req2->next_unhandled == req) {
            req2->next_unhandled = req->next_unhandled;
            break;
        }
    }
    req->next_unhandled = NULL;

    if (!(req->helper = GetHSUser(hs, user->handle_info))) {
        helpserv_notice(user, HSMSG_WTF_WHO_ARE_YOU, hs->helpserv->nick);
        log(HS_LOG, LOG_ERROR, "%s(): Unable to locate %s[%s] on the %s userlist.\n", __FUNCTION__, user->nick, user->handle_info->handle, hs->helpserv->nick);
        return 0;
    }

    if (old_helper) {
        helpserv_notice(user, HSMSG_REQ_REASSIGNED, req->id, old_helper->handle->handle);
        req->helper->reassigned_to[0]++;
        req->helper->reassigned_to[1]++;
        old_helper->reassigned_from[0]++;
        old_helper->reassigned_from[1]++;
    } else {
        helpserv_notice(user, HSMSG_REQ_ASSIGNED_YOU, req->id);
        req->helper->picked_up[0]++;
        req->helper->picked_up[1]++;
    }
    helpserv_show(from_opserv, hs, user, req);
    if (req->user) {
        helpserv_message(hs, req->user, MSGTYPE_REQ_ASSIGNED);
        if (old_helper) {
            helpserv_notice(req->user, HSMSG_REQ_ASSIGNED_AGAIN, req->id, user->handle_info->handle, user->nick);
        } else {
            helpserv_notice(req->user, HSMSG_REQ_ASSIGNED, req->id, user->handle_info->handle, user->nick);
        }
    }
    helpserv_page(PGSRC_STATUS, HSMSG_PAGE_ASSIGN_REQUEST, req->id, req->user ? req->user->nick : "an offline user", req->handle ? req->handle->handle : "not authed", user->nick);
    return 1;
}

static HELPSERV_FUNC(cmd_next) {
    struct helpserv_request *req;

    (void)argc; (void)argv;

    if (!(req = hs->unhandled)) {
        helpserv_notice(user, HSMSG_REQ_NO_UNASSIGNED);
        return 0;
    }
    return helpserv_assign(from_opserv, hs, user, user, req);
}

static HELPSERV_FUNC(cmd_show) {
    struct helpserv_request *req;

    REQUIRE_PARMS(2);

    if (!(req = dict_find(hs->requests, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_REQ_INVALID, argv[1]);
        return 0;
    }

    helpserv_notice(user, HSMSG_REQ_INFO_1, req->id);
    helpserv_show(from_opserv, hs, user, req);
    return 1;
}

static HELPSERV_FUNC(cmd_pickup) {
    struct helpserv_request *req;

    REQUIRE_PARMS(2);

    if (!(req = dict_find(hs->requests, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_REQ_INVALID, argv[1]);
        return 0;
    }
    return helpserv_assign(from_opserv, hs, user, user, req);
}

static HELPSERV_FUNC(cmd_reassign) {
    struct helpserv_request *req;
    struct helpserv_user *target, *hs_user;
    struct userNode *targetuser;

    REQUIRE_PARMS(3);

    if (!(req = dict_find(hs->requests, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_REQ_INVALID, argv[1]);
        return 0;
    }

    if (!(targetuser = GetUserH(argv[2]))) {
        helpserv_notice(user, MSG_NICK_UNKNOWN, argv[2]);
        return 0;
    }

    if (!targetuser->handle_info) {
        helpserv_notice(user, MSG_USER_AUTHENTICATE, targetuser->nick);
        return 0;
    }

    if (!(target = GetHSUser(hs, targetuser->handle_info))) {
        helpserv_notice(user, HSMSG_NOT_IN_USERLIST, targetuser->nick, hs->helpserv->nick);
        return 0;
    }

    hs_user = GetHSUser(hs, user->handle_info);
    if ((hs->persist_types[PERSIST_T_HELPER] == PERSIST_PART) && !GetUserMode(hs->helpchan, user) && (hs_user->level < HlManager)) {
        helpserv_notice(user, HSMSG_REQ_NOTIN_HELPCHAN, targetuser->nick, "be assigned", "they", hs->helpchan->name);
        return 0;
    }

    helpserv_assign(from_opserv, hs, targetuser, user, req);
    return 1;
}

static HELPSERV_FUNC(cmd_addnote) {
    char text[MAX_LINE_SIZE], timestr[MAX_LINE_SIZE];
    struct helpserv_request *req;

    REQUIRE_PARMS(3);

    if (!(req = dict_find(hs->requests, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_REQ_INVALID, argv[1]);
        return 0;
    }

    unsplit_string(argv+2, argc-2);

    strftime(timestr, MAX_LINE_SIZE, HSFMT_TIME, localtime(&now));
    snprintf(text, MAX_LINE_SIZE, HSMSG_REQMSG_HELPER_NOTE, timestr);
    string_list_append(req->text, strdup(text));
    snprintf(text, MAX_LINE_SIZE, HSMSG_REQMSG_HELPER_NOTE2, user->handle_info->handle, argv[2]);
    string_list_append(req->text, strdup(text));

    helpserv_notice(user, HSMSG_REQMSG_NOTE_ADDED, req->id);

    return 1;
}

static HELPSERV_FUNC(cmd_page) {
    REQUIRE_PARMS(2);

    unsplit_string(argv+1, argc-1);
    helpserv_page(PGSRC_COMMAND, HSMSG_PAGE_REQUEST, user->nick, argv[1]);

    return 1;
}

static HELPSERV_FUNC(cmd_stats) {
    struct helpserv_user *target, *hs_user;
    struct handle_info *target_handle;
    struct helpfile_table tbl;
    int i;
    char weekstr[32]; /* "September 10" should be the longest, but with other locales, who knows */
    char intervalstr[INTERVALLEN];
    char buf[12];
    struct tm *week_tm;

    if (!from_opserv) {
        hs_user = GetHSUser(hs, user->handle_info);
    } else {
        hs_user = NULL;
    }

    if (argc > 1) {
        if (!from_opserv) {
            if (hs_user->level < HlManager) {
                helpserv_notice(user, HSMSG_NEED_MANAGER);
                return 0;
            }
        }

        if (!(target_handle = helpserv_get_handle_info(user, argv[1]))) {
            return 0;
        }

        if (!(target = GetHSUser(hs, target_handle))) {
            helpserv_notice(user, HSMSG_NOT_IN_USERLIST, target_handle->handle, hs->helpserv->nick);
            return 0;
        }
    } else {
        if (from_opserv) {
            helpserv_notice(user, HSMSG_OPSERV_NEED_USER);
            return 0;
        }
        target = hs_user;
    }

    helpserv_notice(user, HSMSG_STATS_TOP, hs->helpserv->nick, target->handle->handle);

    tbl.length = 5;
    tbl.width = 2;
    tbl.flags = TABLE_NO_FREE;
    tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
    tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[0][0] = "Week of";
    tbl.contents[0][1] = "Recorded time";
    week_tm = localtime(&last_stats_update);
    week_tm->tm_mday -= week_tm->tm_wday; /* Ensure it'll start on Sunday */
    for (i=0; i < 4; i++) {
        unsigned int week_time = target->time_per_week[i];
        tbl.contents[i+1] = alloca(tbl.width * sizeof(**tbl.contents));
        mktime(week_tm); /* To fix up other stuff after messing with tm_mday */
        if (!strftime(weekstr, 32, "%B %d", week_tm)) {
            tbl.contents[i+1][0] = strdup("(Error)");
        } else {
            tbl.contents[i+1][0] = strdup(weekstr);
        }
        if ((i == 0) && target->join_time) week_time += now - target->join_time;
        intervalString(intervalstr, week_time);
        tbl.contents[i+1][1] = strdup(intervalstr);
        week_tm->tm_mday -= 7;
    }

    helpserv_notice(user, HSMSG_STATS_TIME, hs->helpchan->name);
    table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);

    for (i=1; i < 5; i++) {
        free((char *)tbl.contents[i][0]);
        free((char *)tbl.contents[i][1]);
    }

    tbl.length = 5;
    tbl.width = 3;
    tbl.flags = TABLE_NO_FREE;
    tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
    tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[1] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[2] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[3] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[4] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[0][0] = "Category";
    tbl.contents[0][1] = "This week";
    tbl.contents[0][2] = "Total";

    tbl.contents[1][0] = "Requests picked up";
    for (i=0; i < 2; i++) {
        sprintf(buf, "%u", target->picked_up[i]);
        tbl.contents[1][i+1] = strdup(buf);
    }
    tbl.contents[2][0] = "Requests closed";
    for (i=0; i < 2; i++) {
        sprintf(buf, "%u", target->closed[i]);
        tbl.contents[2][i+1] = strdup(buf);
    }
    tbl.contents[3][0] = "Reassigned from";
    for (i=0; i < 2; i++) {
        sprintf(buf, "%u", target->reassigned_from[i]);
        tbl.contents[3][i+1] = strdup(buf);
    }
    tbl.contents[4][0] = "Reassigned to";
    for (i=0; i < 2; i++) {
        sprintf(buf, "%u", target->reassigned_to[i]);
        tbl.contents[4][i+1] = strdup(buf);
    }

    helpserv_notice(user, HSMSG_STATS_REQS);
    table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);

    for (i=1; i < 5; i++) {
        free((char *)tbl.contents[i][1]);
        free((char *)tbl.contents[i][2]);
    }

    return 1;
}

static HELPSERV_FUNC(cmd_statsreport) {
    int use_privmsg=1;
    struct helpfile_table tbl;
    dict_iterator_t it;
    unsigned int line, i;
    char weekstr[MAX_LINE_SIZE];
    struct tm week_tm;

    if (argc > 1) {
        if (!strcasecmp(argv[1], "NOTICE")) {
            use_privmsg = 0;
        }
    }

    tbl.length = dict_size(hs->users)+1;
    tbl.width = 2;
    tbl.flags = TABLE_NO_FREE;
    tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
    tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[0][0] = "Handle";
    tbl.contents[0][1] = "Time helping";

    for (it=dict_first(hs->users), line=0; it; it=iter_next(it)) {
        struct helpserv_user *hs_user=iter_data(it);

        tbl.contents[++line] = alloca(tbl.width * sizeof(**tbl.contents));
        tbl.contents[line][0] = hs_user->handle->handle;
        tbl.contents[line][1] = malloc(INTERVALLEN);
    }

    localtime_r(&last_stats_update, &week_tm);
    week_tm.tm_mday -= week_tm.tm_wday; /* Ensure it'll start on Sunday */
    week_tm.tm_mday -= 7 * 3; /* Oldest to newest */
    /* 4 to 1 because it's unsigned */
    for (i=4; i > 0; i--, week_tm.tm_mday += 7) {
        for (it=dict_first(hs->users), line=0; it; it=iter_next(it)) {
            struct helpserv_user *hs_user=iter_data(it);
            unsigned int week_time = hs_user->time_per_week[i-1];
            if ((i==1) && hs_user->join_time) week_time += now - hs_user->join_time;
            intervalString((char *)tbl.contents[++line][1], week_time);
        }
        mktime(&week_tm); /* To fix up stuff after messing with tm_mday */
        strftime(weekstr, MAX_LINE_SIZE, HSFMT_STATS_REPORT_WEEK, &week_tm);
        send_target_message(use_privmsg, user->numeric, hs->helpserv, "%s", weekstr);
        table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, use_privmsg ? irc_privmsg : irc_notice, tbl);
    }

    for (line=1; line <= dict_size(hs->users); line++) {
        free((char *)tbl.contents[line][1]);
    }

    return 1;
}

static int
helpserv_in_channel(struct helpserv_bot *hs, struct chanNode *channel) {
    enum page_source pgsrc;
    if (channel == hs->helpchan) return 1;
    for (pgsrc=0; pgsrc<PGSRC_COUNT; pgsrc++) {
        if (channel == hs->page_targets[pgsrc]) return 1;
    }
    return 0;
}

static HELPSERV_FUNC(cmd_move) {
    if (!hs) {
        helpserv_notice(user, HSMSG_INVALID_BOT);
        return 0;
    }

    REQUIRE_PARMS(2);

    if (IsNickName(argv[1])) {
        char *newnick = argv[1], *oldnick, *reason;

        oldnick = alloca(strlen(hs->helpserv->nick) + 1);
        strcpy(oldnick, hs->helpserv->nick);

        if (GetUserH(newnick)) {
            helpserv_notice(user, HSMSG_NICK_EXISTS, newnick);
            return 0;
        }

        dict_remove(helpserv_bots_dict, hs->helpserv->nick);
        NickChange(hs->helpserv, newnick, 0);
        dict_insert(helpserv_bots_dict, hs->helpserv->nick, hs);

        helpserv_notice(user, HSMSG_RENAMED, oldnick, newnick);

        reason = alloca(strlen(user->nick) + strlen(newnick) + strlen(hs->helpchan->name) + strlen(oldnick) + 30);
        sprintf(reason, "HelpServ bot %s (in %s) renamed to %s by %s.", oldnick, hs->helpchan->name, newnick, user->nick);
        global_message(MESSAGE_RECIPIENT_OPERS, reason);

        return 1;
    } else if (IsChannelName(argv[1])) {
        struct chanNode *old_helpchan = hs->helpchan;
        char *newchan = argv[1], *oldchan, *reason;
        struct helpserv_botlist *botlist;

        oldchan = alloca(strlen(hs->helpchan->name) + 1);
        strcpy(oldchan, hs->helpchan->name);

        if (opserv_bad_channel(newchan)) {
            helpserv_notice(user, HSMSG_ILLEGAL_CHANNEL, newchan);
            return 0;
        }

        botlist = dict_find(helpserv_bots_bychan_dict, hs->helpchan->name, NULL);
        helpserv_botlist_remove(botlist, hs);
        if (botlist->used == 0) {
            dict_remove(helpserv_bots_bychan_dict, hs->helpchan->name);
        }

        reason = alloca(strlen(user->nick) + strlen(newchan) + strlen(oldchan) + strlen(hs->helpserv->nick) + 28);
        hs->helpchan = NULL;
        if (!helpserv_in_channel(hs, old_helpchan)) {
            sprintf(reason, "Moved to %s by %s.", newchan, user->nick);
            DelChannelUser(hs->helpserv, old_helpchan, reason, 0);
        }

        if (!(hs->helpchan = GetChannel(newchan))) {
            hs->helpchan = AddChannel(newchan, now, NULL, NULL, 0, hs->helpserv->numeric, NULL);
            AddChannelOp(1, &hs->helpserv, hs->helpchan, hs->helpserv, 0);
        } else if (!helpserv_in_channel(hs, old_helpchan)) {
            AddChannelUser(hs->helpserv, hs->helpchan);
            AddChannelOp(1, &hs->helpserv, hs->helpchan, hs->helpserv, 1);
        }

        if (!(botlist = dict_find(helpserv_bots_bychan_dict, hs->helpchan->name, NULL))) {
            botlist = helpserv_botlist_alloc();
            dict_insert(helpserv_bots_bychan_dict, hs->helpchan->name, botlist);
        }
        helpserv_botlist_append(botlist, hs);

        sprintf(reason, "HelpServ %s (%s) moved to %s by %s.", hs->helpserv->nick, oldchan, newchan, user->nick);
        global_message(MESSAGE_RECIPIENT_OPERS, reason);

        return 1;
    } else {
        helpserv_notice(user, HSMSG_INVALID_MOVE, argv[1]);
        return 0;
    }
}

static HELPSERV_FUNC(cmd_bots) {
    dict_iterator_t it;
    struct helpfile_table tbl;
    unsigned int i;

    (void)argc; (void)argv;

    helpserv_notice(user, HSMSG_BOTLIST_HEADER);

    tbl.length = dict_size(helpserv_bots_dict)+1;
    tbl.width = 3;
    tbl.flags = TABLE_NO_FREE;
    tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
    tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
    tbl.contents[0][0] = "Bot";
    tbl.contents[0][1] = "Channel";
    tbl.contents[0][2] = "Owner";

    for (it=dict_first(helpserv_bots_dict), i=1; it; it=iter_next(it), i++) {
        dict_iterator_t it2;
        struct helpserv_bot *bot;
        struct helpserv_user *owner=NULL;

        bot = iter_data(it);
        
        for (it2=dict_first(bot->users); it2; it2=iter_next(it2)) {
            if (((struct helpserv_user *)iter_data(it2))->level == HlOwner) {
                owner = iter_data(it2);
                break;
            }
        }

        tbl.contents[i] = alloca(tbl.width * sizeof(**tbl.contents));
        tbl.contents[i][0] = iter_key(it);
        tbl.contents[i][1] = bot->helpchan->name;
        tbl.contents[i][2] = owner ? owner->handle->handle : "None";
    }

    table_send((from_opserv ? opserv : hs->helpserv), user->numeric, 0, NULL, tbl);

    return 1;
}

static void helpserv_page_helper_gone(struct helpserv_bot *hs, struct helpserv_request *req, const char *reason) {
    const int from_opserv = 0;

    if (!req->helper) {
        return;
    }

    /* Let the user know that their request is now unhandled */
    if (req->user) {
        helpserv_notice(req->user, HSMSG_REQ_UNASSIGNED, req->id, reason);
    }

    /* Let the other helpers know that the request is now unhandled */
    helpserv_page(PGSRC_STATUS, HSMSG_PAGE_HELPER_GONE, req->id, (req->user ? req->user->nick : "an offline user"), (req->handle ? req->handle->handle : "not authed"), req->helper->handle->handle, reason);

    /* Now put it back in the queue */
    if (hs->unhandled == NULL) {
        /* Nothing there, put it at the front */
        hs->unhandled = req;
        req->next_unhandled = NULL;
    } else {
        /* Should it be at the front? */
        if (hs->unhandled->opened >= req->opened) {
            req->next_unhandled = hs->unhandled;
            hs->unhandled = req;
        } else {
            struct helpserv_request *unhandled;
            /* Find the request that this should be inserted AFTER */
            for (unhandled=hs->unhandled; unhandled->next_unhandled && (unhandled->next_unhandled->opened < req->opened); unhandled = unhandled->next_unhandled);
            req->next_unhandled = unhandled->next_unhandled;
            unhandled->next_unhandled = req;
        }
    }

    req->helper = NULL;
}

/* This takes care of WHINE_DELAY and IDLE_DELAY */
static void run_whine_interval(void *data) {
    struct helpserv_bot *hs=data;
    struct helpfile_table tbl;
    unsigned int i;

    /* First, run the WHINE_DELAY */
    if (hs->intervals[INTERVAL_WHINE_DELAY]
        && (hs->page_types[PGSRC_ALERT] != PAGE_NONE)
        && (hs->page_targets[PGSRC_ALERT] != NULL)) {
        struct helpserv_request *unh;
        struct helpserv_reqlist reqlist;

        helpserv_reqlist_init(&reqlist);

        for (unh = hs->unhandled; unh; unh = unh->next_unhandled) {
            if ((now - unh->opened) >= TIME_T_CAST hs->intervals[INTERVAL_WHINE_DELAY]) {
                helpserv_reqlist_append(&reqlist, unh);
            }
        }

        if (reqlist.used) {
            char strwhinedelay[INTERVALLEN];

            intervalString(strwhinedelay, TIME_T_CAST hs->intervals[INTERVAL_WHINE_DELAY]);
#if ANNOYING_ALERT_PAGES
            tbl.length = reqlist.used + 1;
            tbl.width = 4;
            tbl.flags = TABLE_NO_FREE;
            tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
            tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
            tbl.contents[0][0] = "ID#";
            tbl.contents[0][1] = "Nick";
            tbl.contents[0][2] = "Account";
            tbl.contents[0][3] = "Waiting time";

            for (i=1; i <= reqlist.used; i++) {
                char reqid[12], unh_time[INTERVALLEN];
                unh = reqlist.list[i-1];

                tbl.contents[i] = alloca(tbl.width * sizeof(**tbl.contents));
                sprintf(reqid, "%lu", unh->id);
                tbl.contents[i][0] = strdup(reqid);
                tbl.contents[i][1] = unh->user ? unh->user->nick : "Not online";
                tbl.contents[i][2] = unh->handle ? unh->handle->handle : "Not authed";
                intervalString(unh_time, now - unh->opened);
                tbl.contents[i][3] = strdup(unh_time);
            }

            helpserv_page(PGSRC_ALERT, HSMSG_PAGE_WHINE_HEADER, reqlist.used, strwhinedelay);
            table_send(hs->helpserv, hs->page_targets[PGSRC_ALERT]->name, 0, page_type_funcs[hs->page_types[PGSRC_ALERT]], tbl);

            for (i=1; i <= reqlist.used; i++) {
                free((char *)tbl.contents[i][0]);
                free((char *)tbl.contents[i][3]);
            }
#else
            helpserv_page(PGSRC_ALERT, HSMSG_PAGE_WHINE_HEADER, reqlist.used, strwhinedelay);
#endif
        }

        helpserv_reqlist_clean(&reqlist);
    }

    /* Next run IDLE_DELAY */
    if (hs->intervals[INTERVAL_IDLE_DELAY]
        && (hs->page_types[PGSRC_STATUS] != PAGE_NONE)
        && (hs->page_targets[PGSRC_STATUS] != NULL)) {
        struct modeList mode_list;

        modeList_init(&mode_list);

        for (i=0; i < hs->helpchan->members.used; i++) {
            struct modeNode *mn = hs->helpchan->members.list[i];
            /* Ignore ops. Perhaps this should be a set option? */
            if (mn->modes & MODE_CHANOP) continue;
            /* Check if they've been idle long enough */
            if ((unsigned)(now - mn->idle_since) < hs->intervals[INTERVAL_IDLE_DELAY]) continue;
            /* Add them to the list of idle people.. */
            modeList_append(&mode_list, mn);
        }

        if (mode_list.used) {
            char stridledelay[INTERVALLEN];

            tbl.length = mode_list.used + 1;
            tbl.width = 4;
            tbl.flags = TABLE_NO_FREE;
            tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
            tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
            tbl.contents[0][0] = "Nick";
            tbl.contents[0][1] = "Account";
            tbl.contents[0][2] = "ID#";
            tbl.contents[0][3] = "Idle time";

            for (i=1; i <= mode_list.used; i++) {
                char reqid[12], idle_time[INTERVALLEN];
                struct helpserv_reqlist *reqlist;
                struct modeNode *mn = mode_list.list[i-1];

                tbl.contents[i] = alloca(tbl.width * sizeof(**tbl.contents));
                tbl.contents[i][0] = mn->user->nick;
                tbl.contents[i][1] = mn->user->handle_info ? mn->user->handle_info->handle : "Not authed";

                if ((reqlist = dict_find(helpserv_reqs_bynick_dict, mn->user->nick, NULL))) {
                    int j;

                    for (j = reqlist->used-1; j >= 0; j--) {
                        struct helpserv_request *req = reqlist->list[j];

                        if (req->hs == hs) {
                            sprintf(reqid, "%lu", req->id);
                            break;
                        }
                    }

                    if (j < 0) strcpy(reqid, "None");
                } else {
                    strcpy(reqid, "None");
                }
                tbl.contents[i][2] = strdup(reqid);

                intervalString(idle_time, now - mn->idle_since);
                tbl.contents[i][3] = strdup(idle_time);
            }

            intervalString(stridledelay, TIME_T_CAST(hs->intervals[INTERVAL_IDLE_DELAY]));
            helpserv_page(PGSRC_STATUS, HSMSG_PAGE_IDLE_HEADER, mode_list.used, hs->helpchan->name, stridledelay);
            table_send(hs->helpserv, hs->page_targets[PGSRC_STATUS]->name, 0, page_type_funcs[hs->page_types[PGSRC_STATUS]], tbl);

            for (i=1; i <= mode_list.used; i++) {
                free((char *)tbl.contents[i][2]);
                free((char *)tbl.contents[i][3]);
            }
        }

        modeList_clean(&mode_list);
    }

    if (hs->intervals[INTERVAL_WHINE_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_WHINE_INTERVAL], run_whine_interval, hs);
    }
}

static void run_empty_interval(void *data) {
    struct helpserv_bot *hs=data;
    dict_iterator_t it;

    for (it=dict_first(hs->users); it; it=iter_next(it)) {
        struct helpserv_user *hs_user=iter_data(it);

        if (find_handle_in_channel(hs->helpchan, hs_user->handle, NULL)) {
            return;
        }
    }

    helpserv_page(PGSRC_ALERT, HSMSG_PAGE_EMPTYALERT, hs->helpchan->name);

    if (hs->intervals[INTERVAL_EMPTY_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_EMPTY_INTERVAL], run_empty_interval, hs);
    }
}

static void free_user(void *data) {
    struct helpserv_user *hs_user = data;
    struct helpserv_bot *hs = hs_user->hs;
    struct helpserv_userlist *userlist;
    dict_iterator_t it;

    if (hs->requests) {
        for (it=dict_first(hs->requests); it; it=iter_next(it)) {
            struct helpserv_request *req = iter_data(it);

            if (req->helper == hs_user) {
                helpserv_page_helper_gone(hs, req, "been deleted");
            }
        }
    }

    userlist = dict_find(helpserv_users_byhand_dict, hs_user->handle->handle, NULL);
    if (userlist->used == 1) {
        dict_remove(helpserv_users_byhand_dict, hs_user->handle->handle);
    } else {
        helpserv_userlist_remove(userlist, hs_user);
    }

    free(data);
}

static struct helpserv_bot *register_helpserv(const char *nick, const char *help_channel, const char *registrar) {
    struct helpserv_bot *hs;
    struct helpserv_botlist *botlist;

    /* Laziness dictates calloc, since there's a lot to set to NULL or 0, and
     * it's a harmless default */
    hs = calloc(1, sizeof(struct helpserv_bot));

    hs->helpserv = AddService(nick, helpserv_conf.description);
    reg_privmsg_func(hs->helpserv->num_local, helpserv_botmsg);

    if (!(hs->helpchan = GetChannel(help_channel))) {
        hs->helpchan = AddChannel(help_channel, now, NULL, NULL, 0, hs->helpserv->numeric, NULL);
        AddChannelOp(1, &hs->helpserv, hs->helpchan, hs->helpserv, 0);
    } else {
        AddChannelUser(hs->helpserv, hs->helpchan);
        AddChannelOp(1, &hs->helpserv, hs->helpchan, hs->helpserv, 1);
    }

    if (registrar) hs->registrar = strdup(registrar);

    hs->users = dict_new();
    /* Don't free keys - they use the handle_info's handle field */
    dict_set_free_data(hs->users, free_user);
    hs->requests = dict_new();
    dict_set_free_keys(hs->requests, free);
    dict_set_free_data(hs->requests, free_request);

    dict_insert(helpserv_bots_dict, hs->helpserv->nick, hs);

    if (!(botlist = dict_find(helpserv_bots_bychan_dict, hs->helpchan->name, NULL))) {
        botlist = helpserv_botlist_alloc();
        dict_insert(helpserv_bots_bychan_dict, hs->helpchan->name, botlist);
    }
    helpserv_botlist_append(botlist, hs);

    return hs;
}

static HELPSERV_FUNC(cmd_register) {
    char *nick, *helpchan, *reason;
    struct handle_info *handle;

    if (hs) {
        helpserv_notice(user, HSMSG_ALREADY_REGISTERED, hs->helpserv->nick);
        return 0;
    }
    REQUIRE_PARMS(4);
    nick = argv[1];
    if (!IsNickName(nick)) {
        helpserv_notice(user, HSMSG_ILLEGAL_NICK, nick);
        return 0;
    }
    if (GetUserH(nick)) {
        helpserv_notice(user, HSMSG_NICK_EXISTS, nick);
        return 0;
    }
    helpchan = argv[2];
    if (!IsChannelName(helpchan)) {
        helpserv_notice(user, HSMSG_ILLEGAL_CHANNEL, helpchan);
        HELPSERV_SYNTAX();
        return 0;
    }
    if (opserv_bad_channel(helpchan)) {
        helpserv_notice(user, HSMSG_ILLEGAL_CHANNEL, helpchan);
        return 0;
    }
    if (!(handle = helpserv_get_handle_info(user, argv[3]))) return 0;

    hs = register_helpserv(nick, helpchan, user->handle_info->handle);
    hs->registered = now;
    helpserv_add_user(hs, handle, HlOwner);

    helpserv_notice(user, HSMSG_REG_SUCCESS, handle->handle, nick);

    reason = alloca(strlen(user->nick) + strlen(nick) + strlen(hs->helpchan->name) + 43);
    sprintf(reason, "HelpServ %s (%s) registered to %s by %s.", nick, hs->helpchan->name, handle->handle, user->nick);
    /* Not sent to helpers, since they can't register HelpServ */
    global_message(MESSAGE_RECIPIENT_OPERS, reason);
    return 1;
}

static void unregister_helpserv(struct helpserv_bot *hs) {
    enum message_type msgtype;

    timeq_del(0, NULL, hs, TIMEQ_IGNORE_WHEN|TIMEQ_IGNORE_FUNC);

    /* Requests before users so that it doesn't spam mentioning now-unhandled
     * requests because the users were deleted */
    dict_delete(hs->requests);
    hs->requests = NULL; /* so we don't try to look up requests in free_user() */
    dict_delete(hs->users);
    if (hs->registrar) free(hs->registrar);

    for (msgtype=0; msgtype<MSGTYPE_COUNT; msgtype++) {
        if (hs->messages[msgtype]) free(hs->messages[msgtype]);
    }
}

static void helpserv_free_bot(void *data) {
    unregister_helpserv(data);
    free(data);
}

static HELPSERV_FUNC(cmd_unregister) {
    int len;
    char reason[MAXLEN], *channame, *botname;
    struct helpserv_botlist *botlist;

    if (!from_opserv) {
        if (argc < 2 || strcmp(argv[1], "CONFIRM")) {
            helpserv_notice(user, HSMSG_NEED_UNREG_CONFIRM);
            return 0;
        }
    }

    botlist = dict_find(helpserv_bots_bychan_dict, hs->helpchan->name, NULL);
    helpserv_botlist_remove(botlist, hs);
    if (botlist->used == 0) {
        dict_remove(helpserv_bots_bychan_dict, hs->helpchan->name);
    }

    len = strlen(hs->helpserv->nick) + 1;
    botname = alloca(len);
    safestrncpy(botname, hs->helpserv->nick, len);
    len = strlen(hs->helpchan->name) + 1;
    channame = alloca(len);
    safestrncpy(channame, hs->helpchan->name, len);

    sprintf(reason, "HelpServ unregistered by %s.", user->nick);
    DelUser(hs->helpserv, NULL, 1, reason);

    dict_remove(helpserv_bots_dict, botname);

    sprintf(reason, "HelpServ %s (%s) unregistered by %s.", botname, channame, user->nick);
    global_message(MESSAGE_RECIPIENT_OPERS, reason);
    return 1;
}

static void set_page_target(struct helpserv_bot *hs, enum page_source idx, const char *target) {
    struct chanNode *new_target, *old_target;

    if (target) {
        if (!IsChannelName(target)) {
            log(HS_LOG, LOG_ERROR, "%s has an invalid page target.\n", hs->helpserv->nick);
            return;
        }
        new_target = GetChannel(target);
        if (!new_target) new_target = AddChannel(target, now, NULL, NULL, 0, hs->helpserv->numeric, NULL);
    } else {
        new_target = NULL;
    }
    if (new_target == hs->page_targets[idx]) return;
    old_target = hs->page_targets[idx];
    hs->page_targets[idx] = NULL;
    if (old_target && !helpserv_in_channel(hs, old_target)) {
        DelChannelUser(hs->helpserv, old_target, "Changing page target.", 0);
    }
    if (new_target && !helpserv_in_channel(hs, new_target)) {
        AddChannelUser(hs->helpserv, new_target);
        AddChannelOp(1, &hs->helpserv, new_target, hs->helpserv, 1);
    }
    hs->page_targets[idx] = new_target;
}

static int opt_page_target(struct userNode *user, struct helpserv_bot *hs, int from_opserv, int argc, unsigned char *argv[], enum page_source idx) {
    int changed = 0;
    if (argc > 0) {
        if (!IsOper(user)) {
            helpserv_notice(user, HSMSG_SET_NEED_OPER);
            return 0;
        }
        if (!strcmp(argv[0], "*")) {
            set_page_target(hs, idx, NULL);
            changed = 1;
        } else if (!IsChannelName(argv[0])) {
            helpserv_notice(user, MSG_NOT_CHANNEL_NAME);
            return 0;
        } else {
            set_page_target(hs, idx, argv[0]);
            changed = 1;
        }
    }
    helpserv_notice(user, HSMSG_PAGE_TARGET_VALUE, page_source_names[idx], hs->page_targets[idx] ? hs->page_targets[idx]->name : "None");
    return changed;
}

static HELPSERV_OPTION(opt_pagetarget_command) {
    return opt_page_target(user, hs, from_opserv, argc, argv, PGSRC_COMMAND);
}

static HELPSERV_OPTION(opt_pagetarget_alert) {
    return opt_page_target(user, hs, from_opserv, argc, argv, PGSRC_ALERT);
}

static HELPSERV_OPTION(opt_pagetarget_status) {
    return opt_page_target(user, hs, from_opserv, argc, argv, PGSRC_STATUS);
}

static enum page_type page_type_from_name(const char *name) {
    enum page_type type;
    for (type=0; type<PAGE_COUNT; type++) {
        if (!strcasecmp(page_type_names[type], name)) return type;
    }
    return PAGE_COUNT;
}

static int opt_page_type(struct userNode *user, struct helpserv_bot *hs, int from_opserv, int argc, unsigned char *argv[], enum page_source idx) {
    enum page_type new_type;
    int changed=0;

    if (argc > 0) {
        new_type = page_type_from_name(argv[0]);
        if (new_type == PAGE_COUNT) {
            helpserv_notice(user, HSMSG_INVALID_OPTION, argv[0]);
            return 0;
        }
        hs->page_types[idx] = new_type;
        changed = 1;
    }
    helpserv_notice(user, HSMSG_PAGE_TYPE_VALUE, page_source_names[idx], page_type_names[hs->page_types[idx]]);
    return changed;
}

static HELPSERV_OPTION(opt_pagetype) {
    return opt_page_type(user, hs, from_opserv, argc, argv, PGSRC_COMMAND);
}

static HELPSERV_OPTION(opt_alert_page_type) {
    return opt_page_type(user, hs, from_opserv, argc, argv, PGSRC_ALERT);
}

static HELPSERV_OPTION(opt_status_page_type) {
    return opt_page_type(user, hs, from_opserv, argc, argv, PGSRC_STATUS);
}

static int opt_message(struct userNode *user, struct helpserv_bot *hs, int from_opserv, int argc, unsigned char *argv[], enum message_type idx) {
    int changed=0;

    if (argc > 0) {
        if (hs->messages[idx]) free(hs->messages[idx]);
        unsplit_string(argv, argc);
        hs->messages[idx] = strcmp(argv[0], "*") ? strdup(argv[0]) : NULL;
        changed = 1;
    }
    helpserv_notice(user, HSMSG_MESSAGE_VALUE, message_type_names[idx], hs->messages[idx] ? hs->messages[idx] : "None");
    return changed;
}

static HELPSERV_OPTION(opt_greeting) {
    return opt_message(user, hs, from_opserv, argc, argv, MSGTYPE_GREETING);
}

static HELPSERV_OPTION(opt_openreq) {
    return opt_message(user, hs, from_opserv, argc, argv, MSGTYPE_REQ_OPENED);
}

static HELPSERV_OPTION(opt_assignreq) {
    return opt_message(user, hs, from_opserv, argc, argv, MSGTYPE_REQ_ASSIGNED);
}

static HELPSERV_OPTION(opt_closereq) {
    return opt_message(user, hs, from_opserv, argc, argv, MSGTYPE_REQ_CLOSED);
}

static int opt_interval(struct userNode *user, struct helpserv_bot *hs, int from_opserv, int argc, unsigned char *argv[], enum interval_type idx, unsigned int min) {
    char buf[INTERVALLEN];
    int changed=0;

    if (argc > 0) {
        unsigned long new_int = ParseInterval(argv[0]);
        if (!new_int && strcmp(argv[0], "0")) {
            helpserv_notice(user, MSG_INVALID_DURATION, argv[0]);
            return 0;
        }
        if (new_int && new_int < min) {
            intervalString(buf, min);
            helpserv_notice(user, HSMSG_INVALID_INTERVAL, interval_type_names[idx], "at least", buf);
            return 0;
        }
        hs->intervals[idx] = new_int;
        changed = 1;
    }
    intervalString(buf, hs->intervals[idx]);
    helpserv_notice(user, HSMSG_INTERVAL_VALUE, interval_type_names[idx], hs->intervals[idx] ? buf : "0 (Disabled)");
    return changed;
}

static HELPSERV_OPTION(opt_idle_delay) {
    return opt_interval(user, hs, from_opserv, argc, argv, INTERVAL_IDLE_DELAY, 60);
}

static HELPSERV_OPTION(opt_whine_delay) {
    return opt_interval(user, hs, from_opserv, argc, argv, INTERVAL_WHINE_DELAY, 60);
}

static HELPSERV_OPTION(opt_whine_interval) {
    unsigned int old_val = hs->intervals[INTERVAL_WHINE_INTERVAL];
    int retval;

    retval = opt_interval(user, hs, from_opserv, argc, argv, INTERVAL_WHINE_INTERVAL, 60);

    if (!old_val && hs->intervals[INTERVAL_WHINE_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_WHINE_INTERVAL], run_whine_interval, hs);
    } else if (old_val && !hs->intervals[INTERVAL_WHINE_INTERVAL]) {
        timeq_del(0, run_whine_interval, hs, TIMEQ_IGNORE_WHEN);
    }

    return retval;
}

static HELPSERV_OPTION(opt_empty_interval) {
    unsigned int old_val = hs->intervals[INTERVAL_EMPTY_INTERVAL];
    int retval;

    retval = opt_interval(user, hs, from_opserv, argc, argv, INTERVAL_EMPTY_INTERVAL, 60);

    if (!old_val && hs->intervals[INTERVAL_EMPTY_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_EMPTY_INTERVAL], run_empty_interval, hs);
    } else if (old_val && !hs->intervals[INTERVAL_EMPTY_INTERVAL]) {
        timeq_del(0, run_empty_interval, hs, TIMEQ_IGNORE_WHEN);
    }

    return retval;
}

static HELPSERV_OPTION(opt_stale_delay) {
    return opt_interval(user, hs, from_opserv, argc, argv, INTERVAL_STALE_DELAY, 60);
}

static enum persistence_length persistence_from_name(const char *name) {
    enum persistence_length pers;
    for (pers=0; pers<PERSIST_COUNT; pers++) {
        if (!strcasecmp(name, persistence_length_names[pers])) return pers;
    }
    return PERSIST_COUNT;
}

static int opt_persist(struct userNode *user, struct helpserv_bot *hs, int from_opserv, int argc, unsigned char *argv[], enum persistence_type idx) {
    int changed=0;

    if (argc > 0) {
        enum persistence_length new_pers = persistence_from_name(argv[0]);
        if (new_pers == PERSIST_COUNT) {
            helpserv_notice(user, HSMSG_INVALID_OPTION, argv[0]);
            return 0;
        }
        hs->persist_types[idx] = new_pers;
        changed = 1;
    }
    helpserv_notice(user, HSMSG_PERSIST_VALUE, persistence_type_names[idx], persistence_length_names[hs->persist_types[idx]]);
    return changed;
}

static HELPSERV_OPTION(opt_request_persistence) {
    return opt_persist(user, hs, from_opserv, argc, argv, PERSIST_T_REQUEST);
}

static HELPSERV_OPTION(opt_helper_persistence) {
    return opt_persist(user, hs, from_opserv, argc, argv, PERSIST_T_HELPER);
}

static enum notification_type notification_from_name(const char *name) {
    enum notification_type notify;
    for (notify=0; notify<NOTIFY_COUNT; notify++) {
        if (!strcasecmp(name, notification_type_names[notify])) return notify;
    }
    return NOTIFY_COUNT;
}

static HELPSERV_OPTION(opt_notification) {
    int changed=0;

    if (argc > 0) {
        enum notification_type new_notify = notification_from_name(argv[0]);
        if (new_notify == NOTIFY_COUNT) {
            helpserv_notice(user, HSMSG_INVALID_OPTION, argv[0]);
            return 0;
        }
        if (!from_opserv && (new_notify == NOTIFY_HANDLE)) {
            helpserv_notice(user, HSMSG_SET_NEED_OPER);
            return 0;
        }
        hs->notify = new_notify;
        changed = 1;
    }
    helpserv_notice(user, HSMSG_STRING_VALUE, "Notification", notification_type_names[hs->notify]);
    return changed;
}

static HELPSERV_FUNC(cmd_set) {
    helpserv_option_func_t *opt;

    if (argc < 2) {
        unsigned int i;
        helpserv_option_func_t *display[] = {
            opt_pagetarget_command, opt_pagetarget_alert, opt_pagetarget_status,
            opt_pagetype, opt_alert_page_type, opt_status_page_type,
            opt_greeting, opt_openreq, opt_assignreq, opt_closereq,
            opt_idle_delay, opt_whine_delay, opt_whine_interval, opt_empty_interval, opt_stale_delay,
            opt_request_persistence, opt_helper_persistence, opt_notification
        };

        helpserv_notice(user, HSMSG_QUEUE_OPTIONS);
        for (i=0; i<ArrayLength(display); i++) {
            display[i](user, hs, from_opserv, 0, argv);
        }
        return 1;
    }

    if (!(opt = dict_find(helpserv_option_dict, argv[1], NULL))) {
        helpserv_notice(user, HSMSG_INVALID_OPTION, argv[1]);
        return 0;
    }

    if ((argc > 2) && !from_opserv) {
        struct helpserv_user *hs_user;

        if (!(hs_user = dict_find(hs->users, user->handle_info->handle, NULL))) {
            helpserv_notice(user, HSMSG_WTF_WHO_ARE_YOU, hs->helpserv->nick);
            return 0;
        }

        if (hs_user->level < HlManager) {
            helpserv_notice(user, HSMSG_NEED_MANAGER);
            return 0;
        }
    }
    return opt(user, hs, from_opserv, argc-2, argv+2);
}

static int user_write_helper(const char *key, void *data, void *extra) {
    dict_t db = alloc_object();
    dict_t stats_db = alloc_object();
    dict_t users_db = extra;
    struct helpserv_user *hs_user = data;
    struct string_list *strlist;
    char str[12];
    int i;

    dict_insert(users_db, key, alloc_record_data_object(db));

    dict_insert(db, KEY_HELPER_LEVEL, alloc_record_data_qstring(helpserv_level2str(hs_user->level)));
    dict_insert(db, KEY_HELPER_HELPMODE, alloc_record_data_qstring((hs_user->help_mode ? "1" : "0")));
    dict_insert(db, KEY_HELPER_STATS, alloc_record_data_object(stats_db));

    /* Stats */
    strlist = alloc_string_list(4);
    for (i=0; i < 4; i++) {
        unsigned int week_time = hs_user->time_per_week[i];
        if ((i==0) && hs_user->join_time) week_time += now - hs_user->join_time;
        sprintf(str, "%u", week_time);
        string_list_append(strlist, strdup(str));
    }
    dict_insert(stats_db, KEY_HELPER_STATS_TIME, alloc_record_data_string_list(strlist));
    strlist = alloc_string_list(2);
    for (i=0; i < 2; i++) {
        sprintf(str, "%u", hs_user->picked_up[i]);
        string_list_append(strlist, strdup(str));
    }
    dict_insert(stats_db, KEY_HELPER_STATS_PICKUP, alloc_record_data_string_list(strlist));
    strlist = alloc_string_list(2);
    for (i=0; i < 2; i++) {
        sprintf(str, "%u", hs_user->closed[i]);
        string_list_append(strlist, strdup(str));
    }
    dict_insert(stats_db, KEY_HELPER_STATS_CLOSE, alloc_record_data_string_list(strlist));
    strlist = alloc_string_list(2);
    for (i=0; i < 2; i++) {
        sprintf(str, "%u", hs_user->reassigned_from[i]);
        string_list_append(strlist, strdup(str));
    }
    dict_insert(stats_db, KEY_HELPER_STATS_REASSIGNFROM, alloc_record_data_string_list(strlist));
    strlist = alloc_string_list(2);
    for (i=0; i < 2; i++) {
        sprintf(str, "%u", hs_user->reassigned_to[i]);
        string_list_append(strlist, strdup(str));
    }
    dict_insert(stats_db, KEY_HELPER_STATS_REASSIGNTO, alloc_record_data_string_list(strlist));

    return 0;
}

static int user_read_helper(const char *key, void *data, void *extra) {
    struct record_data *rd = data;
    struct helpserv_bot *hs = extra;
    struct helpserv_user *hs_user;
    struct handle_info *handle;
    dict_t stats;
    enum helpserv_level level;
    char *str;
    struct string_list *strlist;
    unsigned int i;

    if (rd->type != RECDB_OBJECT || !dict_size(rd->d.object)) {
        log(HS_LOG, LOG_ERROR, "Invalid user %s for %s.\n", key, hs->helpserv->nick);
        return 0;
    }

    if (!(handle = get_handle_info(key))) {
        log(HS_LOG, LOG_ERROR, "Nonexistant account %s for %s.\n", key, hs->helpserv->nick);
        return 0;
    }
    str = database_get_data(rd->d.object, KEY_HELPER_LEVEL, RECDB_QSTRING);
    if (str) {
        level = helpserv_str2level(str);
        if (level == HlNone) {
            log(HS_LOG, LOG_ERROR, "Account %s has invalid level %s.\n", str);
            return 0;
        }
    } else {
        log(HS_LOG, LOG_ERROR, "Account %s has no level field for %s.\n", key, hs->helpserv->nick);
        return 0;
    }

    hs_user = helpserv_add_user(hs, handle, level);

    str = database_get_data(rd->d.object, KEY_HELPER_HELPMODE, RECDB_QSTRING);
    hs_user->help_mode = (str && strtol(str, NULL, 0)) ? 1 : 0;

    /* Stats */
    stats = database_get_data(GET_RECORD_OBJECT(rd), KEY_HELPER_STATS, RECDB_OBJECT);

    if (stats) {
        strlist = database_get_data(stats, KEY_HELPER_STATS_TIME, RECDB_STRING_LIST);
        if (strlist) for (i=0; i < 4 && i < strlist->used; i++) hs_user->time_per_week[i] = strtoul(strlist->list[i], NULL, 0);
        strlist = database_get_data(stats, KEY_HELPER_STATS_PICKUP, RECDB_STRING_LIST);
        if (strlist) for (i=0; i < 2 && i < strlist->used; i++) hs_user->picked_up[i] = strtoul(strlist->list[i], NULL, 0);
        strlist = database_get_data(stats, KEY_HELPER_STATS_CLOSE, RECDB_STRING_LIST);
        if (strlist) for (i=0; i < 2 && i < strlist->used; i++) hs_user->closed[i] = strtoul(strlist->list[i], NULL, 0);
        strlist = database_get_data(stats, KEY_HELPER_STATS_REASSIGNFROM, RECDB_STRING_LIST);
        if (strlist) for (i=0; i < 2 && i < strlist->used; i++) hs_user->reassigned_from[i] = strtoul(strlist->list[i], NULL, 0);
        strlist = database_get_data(stats, KEY_HELPER_STATS_REASSIGNTO, RECDB_STRING_LIST);
        if (strlist) for (i=0; i < 2 && i < strlist->used; i++) hs_user->reassigned_to[i] = strtoul(strlist->list[i], NULL, 0);
    }

    return 0;
}

static int request_write_helper(const char *key, void *data, void *extra) {
    dict_t db = alloc_object();
    dict_t requests_db = extra;
    struct helpserv_request *request = data;
    char str[12];

    if (!request->handle) {
        return 0;
    }

    dict_insert(requests_db, key, alloc_record_data_object(db));

    if (request->helper) dict_insert(db, KEY_REQUEST_HELPER, alloc_record_data_qstring(request->helper->handle->handle));
    dict_insert(db, KEY_REQUEST_HANDLE, alloc_record_data_qstring(request->handle->handle));
    sprintf(str, FMT_TIME_T, request->opened);
    dict_insert(db, KEY_REQUEST_OPENED, alloc_record_data_qstring(str));
    dict_insert(db, KEY_REQUEST_TEXT, alloc_record_data_string_list(string_list_copy(request->text)));

    return 0;
}

static int request_read_helper(const char *key, void *data, void *extra) {
    struct record_data *rd = data;
    struct helpserv_bot *hs = extra;
    struct helpserv_request *request;
    struct string_list *strlist;
    char *str;

    if (rd->type != RECDB_OBJECT || !dict_size(rd->d.object)) {
        log(HS_LOG, LOG_ERROR, "Invalid request %s:%s.\n", hs->helpserv->nick, key);
        return 0;
    }

    request = calloc(1, sizeof(struct helpserv_request));

    request->id = strtoul(key, NULL, 0);
    request->hs = hs;
    request->user = NULL;
    request->parent_nick_list = request->parent_hand_list = NULL;

    str = database_get_data(rd->d.object, KEY_REQUEST_HANDLE, RECDB_QSTRING);
    if (str) {
        if (!(request->handle = get_handle_info(str))) {
            log(HS_LOG, LOG_ERROR, "Request %s:%s has an invalid or nonexistant account.\n", hs->helpserv->nick, key);
            free(request);
            return 0;
        }
        if (!(request->parent_hand_list = dict_find(helpserv_reqs_byhand_dict, request->handle->handle, NULL))) {
            request->parent_hand_list = helpserv_reqlist_alloc();
            dict_insert(helpserv_reqs_byhand_dict, request->handle->handle, request->parent_hand_list);
        }
        helpserv_reqlist_append(request->parent_hand_list, request);
    }
    str = database_get_data(rd->d.object, KEY_REQUEST_OPENED, RECDB_QSTRING);
    if (!str) {
        log(HS_LOG, LOG_ERROR, "Request %s:%s has a nonexistant opening time. Using time(NULL).\n", hs->helpserv->nick, key);
        request->opened = time(NULL);
    } else {
        request->opened = TIME_T_CAST strtoul(str, NULL, 0);
    }

    str = database_get_data(rd->d.object, KEY_REQUEST_HELPER, RECDB_QSTRING);
    if (str && !(request->helper = dict_find(hs->users, str, NULL))) {
        log(HS_LOG, LOG_ERROR, "Request %s:%s has an invalid or nonexistant helper.\n", hs->helpserv->nick, key);
        free(request);
        return 0;
    } else {
        if (!hs->unhandled) {
            request->next_unhandled = NULL;
            hs->unhandled = request;
        } else if (hs->unhandled->opened > request->opened) {
            request->next_unhandled = hs->unhandled;
            hs->unhandled = request;
        } else {
            struct helpserv_request *unh;
            for (unh = hs->unhandled; unh->next_unhandled && (unh->next_unhandled->opened < request->opened); unh = unh->next_unhandled);
            request->next_unhandled = unh->next_unhandled;
            unh->next_unhandled = request;
        }
    }

    strlist = database_get_data(rd->d.object, KEY_REQUEST_TEXT, RECDB_STRING_LIST);
    if (!strlist) {
        log(HS_LOG, LOG_ERROR, "Request %s:%s has no text.\n", hs->helpserv->nick, key);
        free(request);
        return 0;
    }
    request->text = string_list_copy(strlist);

    dict_insert(hs->requests, strdup(key), request);

    return 0;
}

static int helpserv_bot_write(const char *key, void *data, void *extra) {
    dict_t users = alloc_object();
    dict_t requests = alloc_object();
    dict_t db = alloc_object();
    dict_t bots_db = extra;
    const struct helpserv_bot *hs = data;
    char str[12];
    enum page_source pagesrc;
    enum message_type msgtype;
    enum interval_type inttype;
    enum persistence_type persisttype;
    struct string_list *slist;

    dict_insert(bots_db, key, alloc_record_data_object(db));

    dict_insert(db, KEY_HELPERS, alloc_record_data_object(users));
    dict_foreach(hs->users, user_write_helper, users);

    dict_insert(db, KEY_REQUESTS, alloc_record_data_object(requests));
    if (hs->persist_types[PERSIST_T_REQUEST] == PERSIST_CLOSE) {
        dict_foreach(hs->requests, request_write_helper, requests);
    }

    dict_insert(db, KEY_HELP_CHANNEL, alloc_record_data_qstring(hs->helpchan->name));
    slist = alloc_string_list(PGSRC_COUNT);
    for (pagesrc=0; pagesrc<PGSRC_COUNT; pagesrc++) {
        struct chanNode *target = hs->page_targets[pagesrc];
        string_list_append(slist, strdup(target ? target->name : "*"));
    }
    dict_insert(db, KEY_PAGE_DEST, alloc_record_data_string_list(slist));
    for (pagesrc=0; pagesrc<PGSRC_COUNT; pagesrc++) {
        const char *src = page_type_names[hs->page_types[pagesrc]];
        dict_insert(db, page_source_names[pagesrc], alloc_record_data_qstring(src));
    }
    for (msgtype=0; msgtype<MSGTYPE_COUNT; msgtype++) {
        const char *msg = hs->messages[msgtype];
        if (msg) dict_insert(db, message_type_names[msgtype], alloc_record_data_qstring(msg));
    }
    for (inttype=0; inttype<INTERVAL_COUNT; inttype++) {
        if (!hs->intervals[inttype]) continue;
        snprintf(str, sizeof(str), "%lu", hs->intervals[inttype]);
        dict_insert(db, interval_type_names[inttype], alloc_record_data_qstring(str));
    }
    for (persisttype=0; persisttype<PERSIST_T_COUNT; persisttype++) {
        const char *persist = persistence_length_names[hs->persist_types[persisttype]];
        dict_insert(db, persistence_type_names[persisttype], alloc_record_data_qstring(persist));
    }
    dict_insert(db, KEY_NOTIFICATION, alloc_record_data_qstring(notification_type_names[hs->notify]));
    snprintf(str, 12, FMT_TIME_T, hs->registered);
    dict_insert(db, KEY_REGISTERED, alloc_record_data_qstring(str));
    if (hs->registrar) dict_insert(db, KEY_REGISTRAR, alloc_record_data_qstring(hs->registrar));
    snprintf(str, 12, "%lu", hs->id_wrap);
    dict_insert(db, KEY_IDWRAP, alloc_record_data_qstring(str));
    snprintf(str, 12, "%lu", hs->last_requestid);
    dict_insert(db, KEY_LAST_REQUESTID, alloc_record_data_qstring(str));

    return 0;
}

static int helpserv_bot_read(const char *key, void *data, void *extra) {
    struct record_data *br = data, *raw_record;
    struct helpserv_bot *hs;
    char *registrar, *helpchannel_name, *str;
    dict_t users, requests;
    enum page_source pagesrc;
    enum message_type msgtype;
    enum interval_type inttype;
    enum persistence_type persisttype;

    (void)extra;

    users = database_get_data(GET_RECORD_OBJECT(br), KEY_HELPERS, RECDB_OBJECT);
    if (!users) {
        log(HS_LOG, LOG_ERROR, "%s has no users.\n", key);
        return 0;
    }
    helpchannel_name = database_get_data(GET_RECORD_OBJECT(br), KEY_HELP_CHANNEL, RECDB_QSTRING);
    if (!helpchannel_name || !IsChannelName(helpchannel_name)) {
        log(HS_LOG, LOG_ERROR, "%s has an invalid channel name.\n", key);
        return 0;
    }
    registrar = database_get_data(GET_RECORD_OBJECT(br), KEY_REGISTRAR, RECDB_QSTRING);

    hs = register_helpserv(key, helpchannel_name, registrar);

    raw_record = dict_find(GET_RECORD_OBJECT(br), KEY_PAGE_DEST, NULL);
    switch (raw_record ? raw_record->type : RECDB_INVALID) {
    case RECDB_QSTRING:
        set_page_target(hs, PGSRC_COMMAND, GET_RECORD_QSTRING(raw_record));
        pagesrc = PGSRC_COMMAND + 1;
        break;
    case RECDB_STRING_LIST: {
        struct string_list *slist = GET_RECORD_STRING_LIST(raw_record);
        for (pagesrc=0; (pagesrc<slist->used) && (pagesrc<PGSRC_COUNT); pagesrc++) {
            const char *dest = slist->list[pagesrc];
            set_page_target(hs, pagesrc, strcmp(dest, "*") ? dest : NULL);
        }
        break;
    }
    default:
        set_page_target(hs, PGSRC_COMMAND, NULL);
        pagesrc = PGSRC_COMMAND + 1;
        break;
    }
    while (pagesrc < PGSRC_COUNT) {
        set_page_target(hs, pagesrc++, hs->page_targets[PGSRC_COMMAND] ? hs->page_targets[PGSRC_COMMAND]->name : NULL);
    }

    for (pagesrc=0; pagesrc<PGSRC_COUNT; pagesrc++) {
        str = database_get_data(GET_RECORD_OBJECT(br), page_source_names[pagesrc], RECDB_QSTRING);
        hs->page_types[pagesrc] = str ? page_type_from_name(str) : PAGE_NONE;
    }

    for (msgtype=0; msgtype<MSGTYPE_COUNT; msgtype++) {
        str = database_get_data(GET_RECORD_OBJECT(br), message_type_names[msgtype], RECDB_QSTRING);
        hs->messages[msgtype] = str ? strdup(str) : NULL;
    }

    for (inttype=0; inttype<INTERVAL_COUNT; inttype++) {
        str = database_get_data(GET_RECORD_OBJECT(br), interval_type_names[inttype], RECDB_QSTRING);
        hs->intervals[inttype] = str ? ParseInterval(str) : 0;
    }
    if (hs->intervals[INTERVAL_WHINE_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_WHINE_INTERVAL], run_whine_interval, hs);
    }
    if (hs->intervals[INTERVAL_EMPTY_INTERVAL]) {
        timeq_add(now + hs->intervals[INTERVAL_EMPTY_INTERVAL], run_empty_interval, hs);
    }

    for (persisttype=0; persisttype<PERSIST_T_COUNT; persisttype++) {
        str = database_get_data(GET_RECORD_OBJECT(br), persistence_type_names[persisttype], RECDB_QSTRING);
        hs->persist_types[persisttype] = str ? persistence_from_name(str) : PERSIST_QUIT;
    }
    str = database_get_data(GET_RECORD_OBJECT(br), KEY_NOTIFICATION, RECDB_QSTRING);
    hs->notify = str ? notification_from_name(str) : NOTIFY_NONE;
    str = database_get_data(GET_RECORD_OBJECT(br), KEY_REGISTERED, RECDB_QSTRING);
    if (str) hs->registered = TIME_T_CAST strtol(str, NULL, 0);
    str = database_get_data(GET_RECORD_OBJECT(br), KEY_IDWRAP, RECDB_QSTRING);
    if (str) hs->id_wrap = strtoul(str, NULL, 0);
    str = database_get_data(GET_RECORD_OBJECT(br), KEY_LAST_REQUESTID, RECDB_QSTRING);
    if (str) hs->last_requestid = strtoul(str, NULL, 0);

    dict_foreach(users, user_read_helper, hs);

    requests = database_get_data(GET_RECORD_OBJECT(br), KEY_REQUESTS, RECDB_OBJECT);
    if (requests) {
        dict_foreach(requests, request_read_helper, hs);
    }

    return 0;
}

static void helpserv_db_write(void) {
    FILE *f;
    char str[12];

    dict_t db = alloc_database(), subdb = alloc_object();

    dict_foreach(helpserv_bots_dict, helpserv_bot_write, subdb);
    dict_insert(db, KEY_BOTS, alloc_record_data_object(subdb));

    sprintf(str, FMT_TIME_T, last_stats_update);
    dict_insert(db, KEY_LAST_STATS_UPDATE, alloc_record_data_qstring(str));

    /* now write the entire database */
    f = fopen(TEMP_DB_NAME, "w");
    write_database(f, db);
    fclose(f);
    free_database(db);
    if (rename(TEMP_DB_NAME, HELPSERV_DB_NAME) == -1) {
        log(NS_LOG, LOG_ERROR, "unable to rename new db to old: %s\n", strerror(errno));
    }
    lastwrite = now;
}

static void helpserv_timed_write(void *data) {
    (void)data;

    helpserv_db_write();
    log(HS_LOG, LOG_INFO, "Successfully made scheduled backup of HelpServ database.\n");
    timeq_add(now + helpserv_conf.db_backup_frequency, helpserv_timed_write, NULL);
}

static void helpserv_db_read(void) {
    dict_t conf_db, object;
    char *str;

    conf_db = parse_database(HELPSERV_DB_NAME);
    if (!conf_db) return;
    if ((object = database_get_data(conf_db, KEY_BOTS, RECDB_OBJECT))) {
        dict_foreach(object, helpserv_bot_read, NULL);
    }

    str = database_get_data(conf_db, KEY_LAST_STATS_UPDATE, RECDB_QSTRING);
    if (str) last_stats_update = TIME_T_CAST strtol(str, NULL, 0);
    else last_stats_update = now;

    free_database(conf_db);
}

static HELPSERV_FUNC(cmd_write) {
    struct timeval start, stop;

    (void)argc; (void)argv;

    gettimeofday(&start, NULL);
    helpserv_db_write();
    gettimeofday(&stop, NULL);
    stop.tv_sec -= start.tv_sec;
    stop.tv_usec -= start.tv_usec;
    if (stop.tv_usec < 0) {
        stop.tv_sec -= 1;
        stop.tv_usec += 1000000;
    }
    helpserv_notice(user, HSMSG_WRITE_SUCCESS, stop.tv_sec, stop.tv_usec/1000);

    return 1;
}

static void helpserv_conf_read(void) {
    dict_t conf_node;
    const char *str;

    if (!(conf_node = conf_get_data(HELPSERV_CONF_NAME, RECDB_OBJECT))) {
        log(HS_LOG, LOG_ERROR, "config node `%s' is missing or has wrong type\n",
                HELPSERV_CONF_NAME);
        return;
    }

    str = database_get_data(conf_node, KEY_DB_BACKUP_FREQ, RECDB_QSTRING);
    helpserv_conf.db_backup_frequency = str ? ParseInterval(str) : 7200;

    str = database_get_data(conf_node, KEY_DESC, RECDB_QSTRING);
    helpserv_conf.description = str;
}

static int helpserv_command_search(const char *command) {
    return dict_find(helpserv_func_dict, command, NULL) ? 1 : 0;
}

static int helpserv_add_alias(const char *alias, const char *command) {
    struct helpserv_cmd *cmd = dict_find(helpserv_func_dict, command, NULL);

    if (!cmd) return 0;
    cmd->refs++;
    dict_insert(helpserv_func_dict, alias, cmd);
    return 1;
}

static int helpserv_del_alias(const char *alias) {
    return dict_remove(helpserv_func_dict, alias);
}

static struct helpserv_cmd *helpserv_define_func(const char *name, helpserv_func_t *func, enum helpserv_level access, long flags) {
    struct helpserv_cmd *cmd = calloc(1, sizeof(struct helpserv_cmd));

    cmd->access = access;
    cmd->weight = 1.0;
    cmd->func = func;
    cmd->flags = flags;
    cmd->refs = 1;
    dict_insert(helpserv_func_dict, name, cmd);

    return cmd;
}

static void helpserv_unref_cmd(void *data) {
    struct helpserv_cmd *cmd = data;
    if (!--cmd->refs) free(cmd);
}

/* Drop requests that persist until part when a user leaves the chan */
static void handle_part(struct userNode *user, struct chanNode *chan) {
    struct helpserv_botlist *botlist;
    struct helpserv_userlist *userlist;
    const int from_opserv = 0; /* for helpserv_notice */
    unsigned int i;

    if ((botlist = dict_find(helpserv_bots_bychan_dict, chan->name, NULL))) {
        for (i=0; i < botlist->used; i++) {
            struct helpserv_bot *hs = botlist->list[i];
            dict_iterator_t it;

            if (hs->helpserv == NULL) {
                continue;
            }

            for (it=dict_first(hs->requests); it; it=iter_next(it)) {
                struct helpserv_request *req = iter_data(it);

                if (user == req->user) {
                    if (hs->persist_types[PERSIST_T_REQUEST] == PERSIST_PART) {
                        helpserv_message(hs, user, MSGTYPE_REQ_DROPPED);
                        helpserv_notice(user, HSMSG_REQ_DROPPED_PART, chan->name, req->id);

                        if (req->helper && (hs->notify >= NOTIFY_DROP)) helpserv_notify(req->helper, HSMSG_NOTIFY_REQ_DROP, req->id, "they have left the help channel");

                        dict_remove(hs->requests, iter_key(it));
                        break;
                    }
                }
            }
        }
    }
    
    if (user->handle_info && (userlist = dict_find(helpserv_users_byhand_dict, user->handle_info->handle, NULL))) {
        for (i=0; i < userlist->used; i++) {
            struct helpserv_user *hs_user = userlist->list[i];
            struct helpserv_bot *hs = hs_user->hs;
            dict_iterator_t it;

            if ((hs->helpserv == NULL) || (hs->helpchan != chan) || find_handle_in_channel(hs->helpchan, user->handle_info, user)) {
                continue;
            }

            /* In case of the clock being set back for whatever reason,
             * minimize penalty. Don't duplicate this in handle_quit because
             * when users quit, handle_part is called for every channel first.
             */
            if (hs_user->join_time && (hs_user->join_time < now)) {
                hs_user->time_per_week[0] += (unsigned int)(now - hs_user->join_time);
            }
            hs_user->join_time = 0;

            for (it=dict_first(hs->requests); it; it=iter_next(it)) {
                struct helpserv_request *req=iter_data(it);

                if (hs->persist_types[PERSIST_T_HELPER] == PERSIST_PART) {
                    if (req->helper == hs_user) {
                        char *reason;
                        reason = alloca(8 + strlen(chan->name));
                        sprintf(reason, "parted %s", chan->name);
                        helpserv_page_helper_gone(hs, req, reason);
                    }
                }
            }

            if (hs->intervals[INTERVAL_EMPTY_INTERVAL]) {
                timeq_del(0, run_empty_interval, hs, TIMEQ_IGNORE_WHEN);
                run_empty_interval(hs);
            }
        }
    }
}

/* Drop requests that persist until part or quit when a user quits. Otherwise
 * set req->user to null (it's no longer valid) if they have a handle,
 * and drop it if they don't (nowhere to store the request).
 *
 * Unassign requests where req->helper persists until the helper parts or
 * quits. */
static void handle_quit(struct userNode *user) {
    struct helpserv_reqlist *reqlist;
    struct helpserv_userlist *userlist;
    unsigned int i, n;

    if (IsLocal(user)) {
        struct helpserv_bot *hs;
        if ((hs = dict_find(helpserv_bots_dict, user->nick, NULL))) {
            hs->helpserv = NULL;
        }
        return;
    }

    if ((reqlist = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL))) {
        n = reqlist->used;
        for (i=0; i < n; i++) {
            struct helpserv_request *req = reqlist->list[0];

            if ((req->hs->persist_types[PERSIST_T_REQUEST] == PERSIST_QUIT) || !req->handle) {
                char buf[12];
                sprintf(buf, "%lu", req->id);

                if (req->helper && (req->hs->notify >= NOTIFY_DROP)) helpserv_notify(req->helper, HSMSG_NOTIFY_REQ_DROP, req->id, "the user has quit");

                dict_remove(req->hs->requests, buf);
            } else {
                req->user = NULL;
                req->parent_nick_list = NULL;
                helpserv_reqlist_remove(reqlist, req);

                if (req->helper && (req->hs->notify >= NOTIFY_USER)) helpserv_notify(req->helper, HSMSG_NOTIFY_USER_QUIT, req->id, user->nick);
            }
        }

        dict_remove(helpserv_reqs_bynick_dict, user->nick);
    }

    if (user->handle_info && (userlist = dict_find(helpserv_users_byhand_dict, user->handle_info->handle, NULL))) {
        for (i=0; i < userlist->used; i++) {
            struct helpserv_user *hs_user = userlist->list[i];
            struct helpserv_bot *hs = hs_user->hs;
            dict_iterator_t it;

            if ((hs->helpserv == NULL) || user->next_authed || (user->handle_info->users != user)) {
                continue;
            }

            for (it=dict_first(hs->requests); it; it=iter_next(it)) {
                struct helpserv_request *req=iter_data(it);

                if ((hs->persist_types[PERSIST_T_HELPER] == PERSIST_QUIT) && (req->helper == hs_user)) {
                    helpserv_page_helper_gone(hs, req, "disconnected");
                }
            }
        }
    }
}

static void associate_requests_bybot(struct helpserv_bot *hs, struct userNode *user, int force_greet) {
    struct helpserv_reqlist *reqlist, *hand_reqlist=NULL;
    struct helpserv_request *newest=NULL, *nicknewest=NULL;
    unsigned int i;
    const int from_opserv = 0; /* For helpserv_notice */
    
    if (!(user->handle_info && (hand_reqlist = dict_find(helpserv_reqs_byhand_dict, user->handle_info->handle, NULL))) && !force_greet) {
        return;
    }

    reqlist = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL);

    if (hand_reqlist) {
        for (i=0; i < hand_reqlist->used; i++) {
            struct helpserv_request *req=hand_reqlist->list[i];

            if (req->user || (req->hs != hs)) {
                continue;
            }

            req->user = user;
            if (!reqlist) {
                reqlist = helpserv_reqlist_alloc();
                dict_insert(helpserv_reqs_bynick_dict, user->nick, reqlist);
            }
            req->parent_nick_list = reqlist;
            helpserv_reqlist_append(reqlist, req);

            if (req->helper && (hs->notify >= NOTIFY_USER)) helpserv_notify(req->helper, HSMSG_NOTIFY_USER_FOUND, req->id, user->nick);

            if (!newest || (newest->opened < req->opened)) {
                newest = req;
            }
        }
    }

    /* If it's supposed to force a greeting, only bail out if there are no
     * requests at all. If it's not supposed to force a greeting, bail out if
     * nothing was changed. */
    if (!(newest || (force_greet && reqlist))) {
        return;
    }

    /* Possible conditions here:
     * 1. newest == NULL, force_greeting == 1, reqlist != NULL
     * 2. newest != NULL, force_greeting doesn't matter, reqlist != NULL */

    /* Figure out which request will get their next message */
    for (i=0; i < reqlist->used; i++) {
        struct helpserv_request *req=reqlist->list[i];

        if (req->hs != hs) {
            continue;
        }

        if (!nicknewest || (nicknewest->opened < req->opened)) {
            nicknewest = req;
        }
    }

    if ((force_greet && nicknewest) || (nicknewest == newest)) {
        /* Let the user know. Either the user is forced to be greeted, or the
         * above has changed which request will get their next message. */
        helpserv_notice(user, HSMSG_GREET_EXISTING_REQ, hs->helpchan->name, nicknewest->id);
    }
}

static void associate_requests_bychan(struct chanNode *chan, struct userNode *user, int force_greet) {
    struct helpserv_botlist *botlist;
    unsigned int i;

    if (!(botlist = dict_find(helpserv_bots_bychan_dict, chan->name, NULL))) {
        return;
    }

    for (i=0; i < botlist->used; i++) {
        associate_requests_bybot(botlist->list[i], user, force_greet);
    }
}


/* Greet users upon joining a helpserv channel (if greeting is set) and set
 * req->user to the user joining for all requests owned by the user's handle
 * (if any) with a req->user == NULL */
static void handle_join(struct userNode *user, struct chanNode *chan) {
    struct helpserv_botlist *botlist;
    unsigned int i;
    const int from_opserv = 0; /* for helpserv_notice */

    if (IsLocal(user) || self->uplink->burst) {
        return;
    }

    if ((botlist = dict_find(helpserv_bots_bychan_dict, chan->name, NULL))) {
        for (i=0; i < botlist->used; i++) {
            struct helpserv_bot *hs=botlist->list[i];
            helpserv_message(hs, user, MSGTYPE_GREETING);
            associate_requests_bybot(hs, user, 1);

            if (user->handle_info) {
                struct helpserv_user *hs_user;

                if ((hs_user = dict_find(hs->users, user->handle_info->handle, NULL))) {
                    if (!hs_user->join_time) hs_user->join_time = now;
                }
            }
        }
    }
}

/* Update helpserv_reqs_bynick_dict upon nick change */
static void handle_nickchange(struct userNode *user, const char *new_nick) {
    char old_nick[NICKLEN+1];
    struct helpserv_reqlist *reqlist;
    unsigned int i;

    if (!(reqlist = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL))) {
        return;
    }

    /* Don't free the list */
    dict_remove2(helpserv_reqs_bynick_dict, user->nick, 1);

    /* new_nick may be a temporary pointer. The dict stores the key pointer.
     * Make sure it knows about the change but still points to user->nick. */
    strcpy(old_nick, user->nick);
    strcpy(user->nick, new_nick);
    dict_insert(helpserv_reqs_bynick_dict, user->nick, reqlist);
    strcpy(user->nick, old_nick);

    for (i=0; i < reqlist->used; i++) {
        struct helpserv_request *req=reqlist->list[i];

        if (req->helper && (req->hs->notify >= NOTIFY_USER)) helpserv_notify(req->helper, HSMSG_NOTIFY_USER_NICK, req->id, user->nick, new_nick);
    }
}

/* Also update helpserv_reqs_byhand_dict upon handle rename */
static void handle_nickserv_rename(struct handle_info *handle, const char *new_handle) {
    char old_handle[NICKSERV_HANDLE_LEN+1];
    struct helpserv_reqlist *reqlist;
    struct helpserv_userlist *userlist;
    unsigned int i;

    /* First, rename the handle in the requests dict */
    if ((reqlist = dict_find(helpserv_reqs_byhand_dict, handle->handle, NULL))) {
        /* Don't free the list */
        dict_remove2(helpserv_reqs_byhand_dict, handle->handle, 1);

        /* new_handle may be a temporary pointer. The dict stores the key pointer.
         * Make sure it knows about the change but still points to handle->handle,
         * as handle->handle will soon store the new name. */
        strcpy(old_handle, handle->handle);
        strcpy(handle->handle, new_handle);
        dict_insert(helpserv_reqs_byhand_dict, handle->handle, reqlist);
        strcpy(handle->handle, old_handle);
    }

    /* Second, rename the handle in the users dict */
    if ((userlist = dict_find(helpserv_users_byhand_dict, handle->handle, NULL))) {
        dict_remove2(helpserv_users_byhand_dict, handle->handle, 1);

        strcpy(old_handle, handle->handle);
        strcpy(handle->handle, new_handle);
        dict_insert(helpserv_users_byhand_dict, handle->handle, userlist);
        strcpy(handle->handle, old_handle);
    }
    
    if (reqlist) {
        for (i=0; i < reqlist->used; i++) {
            struct helpserv_request *req=reqlist->list[i];

            if (req->helper && (req->hs->notify >= NOTIFY_HANDLE)) helpserv_notify(req->helper, HSMSG_NOTIFY_HAND_RENAME, req->id, handle->handle, new_handle);
        }
    }
}

/* Deals with two cases:
 * 1. No handle -> handle
 *    - Bots with a request assigned to both the user (w/o handle) and the
 *      handle can exist in this case. When a message is sent,
 *      helpserv_usermsg will append it to the most recently opened request.
 *    - Requests assigned to the handle are NOT assigned to the user, since
 *      multiple people can auth to the same handle at once. Wait for them to
 *      join / privmsg before setting req->user.
 * 2. Handle -> handle
 *    - Generally avoided, but sometimes the code may allow this.
 *    - Requests that persist only until part/quit are brought along to the
 *      new handle.
 *    - Requests that persist until closed (stay saved with the handle) are
 *      left with the old handle. This is to prevent the confusing situation
 *      where some requests are carried over to the new handle, and some are
 *      left (because req->handle is the same for all of them, but only some
 *      have req->user set).
 * - In either of the above cases, if a user is on a bot's userlist and has
 *   requests assigned to them, it will give them a list. */
static void handle_nickserv_auth(struct userNode *user, struct handle_info *old_handle) {
    struct helpserv_reqlist *reqlist, *dellist=NULL, *hand_reqlist, *oldhand_reqlist;
    struct helpserv_userlist *userlist;
    unsigned int i, j;
    dict_iterator_t it;
    const int from_opserv = 0; /* for helpserv_notice */

    if (!user->handle_info) return; /* Authed user is quitting */

    if ((userlist = dict_find(helpserv_users_byhand_dict, user->handle_info->handle, NULL))) {
        for (i=0; i < userlist->used; i++) {
            struct helpserv_user *hs_user = userlist->list[i];
            struct helpserv_bot *hs = hs_user->hs;
            struct helpserv_reqlist helper_reqs;
            struct helpfile_table tbl;

            if (!hs_user->join_time && find_handle_in_channel(hs->helpchan, hs_user->handle, NULL)) {
                hs_user->join_time = now;
            }

            helpserv_reqlist_init(&helper_reqs);

            for (it=dict_first(hs->requests); it; it=iter_next(it)) {
                struct helpserv_request *req=iter_data(it);

                if (req->helper == hs_user) {
                    helpserv_reqlist_append(&helper_reqs, req);
                }
            }

            if (helper_reqs.used) {
                tbl.length = helper_reqs.used+1;
                tbl.width = 5;
                tbl.flags = TABLE_NO_FREE;
                tbl.contents = alloca(tbl.length * sizeof(*tbl.contents));
                tbl.contents[0] = alloca(tbl.width * sizeof(**tbl.contents));
                tbl.contents[0][0] = "Bot";
                tbl.contents[0][1] = "ID#";
                tbl.contents[0][2] = "Nick";
                tbl.contents[0][3] = "Account";
                tbl.contents[0][4] = "Opened";

                for (j=1; j <= helper_reqs.used; j++) {
                    struct helpserv_request *req=helper_reqs.list[j-1];
                    char reqid[12], timestr[MAX_LINE_SIZE];

                    tbl.contents[j] = alloca(tbl.width * sizeof(**tbl.contents));
                    tbl.contents[j][0] = req->hs->helpserv->nick;
                    sprintf(reqid, "%lu", req->id);
                    tbl.contents[j][1] = strdup(reqid);
                    tbl.contents[j][2] = req->user ? req->user->nick : "Not online";
                    tbl.contents[j][3] = req->handle ? req->handle->handle : "Not authed";
                    strftime(timestr, MAX_LINE_SIZE, HSFMT_TIME, localtime(&req->opened));
                    tbl.contents[j][4] = strdup(timestr);
                }

                helpserv_notice(user, HSMSG_REQLIST_AUTH);
                table_send(hs->helpserv, user->numeric, 0, NULL, tbl);

                for (j=1; j <= helper_reqs.used; j++) {
                    free((char *)tbl.contents[j][1]);
                    free((char *)tbl.contents[j][4]);
                }
            }

            helpserv_reqlist_clean(&helper_reqs);
        }
    }


    if (!(reqlist = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL))) {
        for (i=0; i < user->channels.used; i++) {
            associate_requests_bychan(user->channels.list[i]->channel, user, 0);
        }
        return;
    }

    if (!(hand_reqlist = dict_find(helpserv_reqs_byhand_dict, user->handle_info->handle, NULL))) {
        hand_reqlist = helpserv_reqlist_alloc();
        dict_insert(helpserv_reqs_byhand_dict, user->handle_info->handle, hand_reqlist);
    }

    if (old_handle) {
        dellist = helpserv_reqlist_alloc();
        oldhand_reqlist = dict_find(helpserv_reqs_byhand_dict, old_handle->handle, NULL);
    } else {
        oldhand_reqlist = NULL;
    }

    for (i=0; i < reqlist->used; i++) {
        struct helpserv_request *req = reqlist->list[i];
        struct helpserv_bot *hs=req->hs;

        if (!old_handle || hs->persist_types[PERSIST_T_REQUEST] == PERSIST_PART || hs->persist_types[PERSIST_T_REQUEST] == PERSIST_QUIT) {
            /* The request needs to be assigned to the new handle; either it
             * only persists until part/quit (so it makes sense to keep it as
             * close to the user as possible, and if it's made persistent later
             * then it's attached to the new handle) or there is no old handle.
             */

            req->handle = user->handle_info;

            req->parent_hand_list = hand_reqlist;
            helpserv_reqlist_append(hand_reqlist, req);

            if (oldhand_reqlist) {
                if (oldhand_reqlist->used == 1) {
                    dict_remove(helpserv_reqs_byhand_dict, old_handle->handle);
                    oldhand_reqlist = NULL;
                } else {
                    helpserv_reqlist_remove(oldhand_reqlist, req);
                }
            }

            if (old_handle) {
                char buf[CHANNELLEN + 14];

                if (hs->persist_types[PERSIST_T_REQUEST] == PERSIST_PART) {
                    sprintf(buf, "part channel %s", hs->helpchan->name);
                } else {
                    strcpy(buf, "quit irc");
                }

                helpserv_notice(user, HSMSG_REQ_AUTH_MOVED, user->handle_info->handle, hs->helpchan->name, req->id, old_handle->handle, buf);
                if (req->helper && (hs->notify >= NOTIFY_HANDLE)) helpserv_notify(req->helper, HSMSG_NOTIFY_HAND_MOVE, req->id, user->handle_info->handle, old_handle->handle);
            } else {
                if (req->helper && (hs->notify >= NOTIFY_HANDLE)) helpserv_notify(req->helper, HSMSG_NOTIFY_HAND_AUTH, req->id, user->nick, user->handle_info->handle);
            }
        } else {
            req->user = NULL;
            req->parent_nick_list = NULL;
            /* Would rather not mess with the list while iterating through
             * it */
            helpserv_reqlist_append(dellist, req);

            helpserv_notice(user, HSMSG_REQ_AUTH_STUCK, user->handle_info->handle, hs->helpchan->name, req->id, old_handle->handle);
            if (req->helper && (hs->notify >= NOTIFY_HANDLE)) helpserv_notify(req->helper, HSMSG_NOTIFY_HAND_STUCK, req->id, user->nick, user->handle_info->handle, old_handle->handle);
        }
    }

    if (old_handle) {
        if (dellist->used) {
            if (dellist->used == reqlist->used) {
                dict_remove(helpserv_reqs_bynick_dict, user->nick);
            } else {
                for (i=0; i < dellist->used; i++) {
                    helpserv_reqlist_remove(reqlist, dellist->list[i]);
                }
            }
        }
        helpserv_reqlist_free(dellist);
    }

    for (i=0; i < user->channels.used; i++) {
        associate_requests_bychan(user->channels.list[i]->channel, user, 0);
    }
}


/* Disassociate all requests from the handle. If any have req->user == NULL
 * then give them to the user doing the unregistration (if not an oper/helper)
 * otherwise the first nick it finds authed (it lets them know about this). If
 * there are no users authed to the handle online, the requests are lost. This
 * may result in the user having >1 request/bot, and messages go to the most
 * recently opened request.
 *
 * Also, remove the user from all bots that it has access in.
 * helpserv_del_user() will take care of unassigning the requests. */
static void handle_nickserv_unreg(struct userNode *user, struct handle_info *handle) {
    struct helpserv_reqlist *hand_reqlist;
    struct helpserv_userlist *userlist;
    unsigned int i, n;
    const int from_opserv = 0; /* for helpserv_notice */
    struct helpserv_bot *hs; /* for helpserv_notice */

    if ((userlist = dict_find(helpserv_users_byhand_dict, handle->handle, NULL))) {
        n=userlist->used;

        /* Each time helpserv_del_user is called, that entry is going to be
         * taken out of userlist... so this should cope with that */
        for (i=0; i < n; i++) {
            struct helpserv_user *hs_user=userlist->list[0];
            struct helpserv_bot *hs=hs_user->hs;

            helpserv_del_user(hs, hs_user);
        }
    }

    if (!(hand_reqlist = dict_find(helpserv_reqs_byhand_dict, handle->handle, NULL))) {
        return;
    }

    n = hand_reqlist->used;
    for (i=0; i < n; i++) {
        struct helpserv_request *req=hand_reqlist->list[0];
        hs = req->hs;

        req->handle = NULL;
        req->parent_hand_list = NULL;
        helpserv_reqlist_remove(hand_reqlist, req);
        if (user && req->helper && (hs->notify >= NOTIFY_HANDLE)) helpserv_notify(req->helper, HSMSG_NOTIFY_HAND_UNREG, req->id, handle->handle, user->nick);

        if (!req->user) {
            if (!user) {
                /* This is probably an expire. Silently remove everything. */

                char buf[12];
                if (req->helper && (hs->notify >= NOTIFY_DROP)) helpserv_notify(req->helper, HSMSG_NOTIFY_REQ_DROP, req->id, "their account has been unregistered and no online users were authenticated to it");
                sprintf(buf, "%lu", req->id);
                dict_remove(req->hs->requests, buf);
            } else if (user->handle_info == handle) {
                req->user = user;
                if (!(req->parent_nick_list = dict_find(helpserv_reqs_bynick_dict, user->nick, NULL))) {
                    req->parent_nick_list = helpserv_reqlist_alloc();
                    dict_insert(helpserv_reqs_bynick_dict, user->nick, req->parent_nick_list);
                }
                helpserv_reqlist_append(req->parent_nick_list, req);

                if (hs->persist_types[PERSIST_T_REQUEST] == PERSIST_CLOSE) {
                    helpserv_notice(req->user, HSMSG_REQ_WARN_UNREG, handle->handle, hs->helpchan->name, req->id);
                }
            } else {
                if (handle->users) {
                    req->user = handle->users;

                    if (!(req->parent_nick_list = dict_find(helpserv_reqs_bynick_dict, req->user->nick, NULL))) {
                        req->parent_nick_list = helpserv_reqlist_alloc();
                        dict_insert(helpserv_reqs_bynick_dict, req->user->nick, req->parent_nick_list);
                    }
                    helpserv_reqlist_append(req->parent_nick_list, req);

                    helpserv_notice(req->user, HSMSG_REQ_ASSIGNED_UNREG, handle->handle, hs->helpchan->name, req->id);
                    if (req->helper && (hs->notify >= NOTIFY_USER)) helpserv_notify(req->helper, HSMSG_NOTIFY_USER_MOVE, req->id, handle->handle, req->user->nick);
                } else {
                    char buf[12];

                    helpserv_notice(user, HSMSG_REQ_DROPPED_UNREG, handle->handle, hs->helpchan->name, req->id);
                    if (req->helper && (hs->notify >= NOTIFY_DROP)) helpserv_notify(req->helper, HSMSG_NOTIFY_REQ_DROP, req->id, "their account has been unregistered and no online users were authenticated to it");

                    sprintf(buf, "%lu", req->id);
                    dict_remove(req->hs->requests, buf);
                }
            }
        } else if (!user) {
            /* This is probably a merge. Move everything over to the new user's handle. */
            /* FIXME: There should be a better way to check for this */

            if (req->user->handle_info) {
                req->handle = req->user->handle_info;

                if (!(req->parent_hand_list = dict_find(helpserv_reqs_byhand_dict, req->handle->handle, NULL))) {
                    req->parent_hand_list = helpserv_reqlist_alloc();
                    dict_insert(helpserv_reqs_byhand_dict, req->handle->handle, req->parent_hand_list);
                }
                helpserv_reqlist_append(req->parent_hand_list, req);
            }
        }
    }

    dict_remove(helpserv_reqs_byhand_dict, handle->handle);
}

static time_t helpserv_next_stats(time_t after_when) {
    struct tm *timeinfo=localtime(&after_when);

    /* This works because mktime(3) says it will accept out-of-range values
     * and fix them for us. tm_wday and tm_yday are ignored. */
    timeinfo->tm_mday += 7 - timeinfo->tm_wday;

    /* We want to run stats at midnight (local time). Perhaps make this
     * configurable? */
    timeinfo->tm_sec = timeinfo->tm_min = timeinfo->tm_hour = 0;

    return mktime(timeinfo);
}

/* If data != NULL, then don't add to the timeq */
static void helpserv_run_stats(void *data) {
    struct helpserv_bot *hs;
    struct helpserv_user *hs_user;
    int i;
    dict_iterator_t it, it2;

    (void)data;

    for (it=dict_first(helpserv_bots_dict); it; it=iter_next(it)) {
        hs = iter_data(it);

        for (it2=dict_first(hs->users); it2; it2=iter_next(it2)) {
            hs_user = iter_data(it2);

            if (hs_user->join_time) {
                hs_user->time_per_week[0] += now - hs_user->join_time;
                hs_user->join_time = now;
            }

            /* Shift time */
            for (i=3; i > 0; i--) hs_user->time_per_week[i] = hs_user->time_per_week[i-1];

            /* Reset it for this week */
            hs_user->time_per_week[0] = hs_user->picked_up[0] = hs_user->closed[0] = hs_user->reassigned_from[0] = hs_user->reassigned_to[0] = 0;
        }
    }

    last_stats_update = now;
    if (!data) {
        timeq_add(helpserv_next_stats(now), helpserv_run_stats, NULL);
    }
}


static void
helpserv_define_option(const char *name, helpserv_option_func_t *func) {
    dict_insert(helpserv_option_dict, name, func);
}

static void helpserv_db_cleanup(void) {
    unreg_part_func(handle_part);
    unreg_del_user_func(handle_quit);
    close_helpfile(helpserv_helpfile);
    dict_delete(helpserv_func_dict);
    dict_delete(helpserv_option_dict);
    dict_delete(helpserv_bots_dict);
    dict_delete(helpserv_bots_bychan_dict);
    dict_delete(helpserv_reqs_bynick_dict);
    dict_delete(helpserv_reqs_byhand_dict);
    dict_delete(helpserv_users_byhand_dict);
}

void init_helpserv() {
    helpserv_enabled = 1;

    helpserv_conf_read();
    conf_register_reload(helpserv_conf_read);

    helpserv_func_dict = dict_new();
    dict_set_free_data(helpserv_func_dict, helpserv_unref_cmd);
    helpserv_define_func("HELP", cmd_help, HlNone, CMD_NOT_OVERRIDE|CMD_IGNORE_EVENT);
    helpserv_define_func("VERSION", cmd_version, HlNone, CMD_NOT_OVERRIDE|CMD_IGNORE_EVENT);
    helpserv_define_func("LIST", cmd_list, HlHelper, CMD_NEED_BOT|CMD_IGNORE_EVENT);
    helpserv_define_func("NEXT", cmd_next, HlHelper, CMD_NEED_BOT|CMD_NEVER_FROM_OPSERV);
    helpserv_define_func("PICKUP", cmd_pickup, HlHelper, CMD_NEED_BOT|CMD_NEVER_FROM_OPSERV);
    helpserv_define_func("REASSIGN", cmd_reassign, HlManager, CMD_NEED_BOT|CMD_NEVER_FROM_OPSERV);
    helpserv_define_func("CLOSE", cmd_close, HlHelper, CMD_NEED_BOT|CMD_NEVER_FROM_OPSERV);
    helpserv_define_func("SHOW", cmd_show, HlHelper, CMD_NEED_BOT|CMD_IGNORE_EVENT);
    helpserv_define_func("ADDNOTE", cmd_addnote, HlHelper, CMD_NEED_BOT);
    helpserv_define_func("ADDOWNER", cmd_addowner, HlOper, CMD_NEED_BOT|CMD_FROM_OPSERV_ONLY);
    helpserv_define_func("DELOWNER", cmd_deluser, HlOper, CMD_NEED_BOT|CMD_FROM_OPSERV_ONLY);
    helpserv_define_func("ADDHELPER", cmd_addhelper, HlManager, CMD_NEED_BOT);
    helpserv_define_func("ADDMANAGER", cmd_addmanager, HlOwner, CMD_NEED_BOT);
    helpserv_define_func("DELUSER", cmd_deluser, HlManager, CMD_NEED_BOT);
    helpserv_define_func("HELPERS", cmd_helpers, HlNone, CMD_NEED_BOT);
    helpserv_define_func("CLVL", cmd_clvl, HlOwner, CMD_NEED_BOT);
    helpserv_define_func("PAGE", cmd_page, HlHelper, CMD_NEED_BOT);
    helpserv_define_func("SET", cmd_set, HlHelper, CMD_NEED_BOT);
    helpserv_define_func("STATS", cmd_stats, HlHelper, CMD_NEED_BOT);
    helpserv_define_func("STATSREPORT", cmd_statsreport, HlManager, CMD_NEED_BOT);
    helpserv_define_func("UNREGISTER", cmd_unregister, HlOwner, CMD_NEED_BOT);
    helpserv_define_func("VERSION", cmd_version, HlNone, CMD_NOT_OVERRIDE|CMD_IGNORE_EVENT);
    helpserv_define_func("READHELP", cmd_readhelp, HlOper, CMD_FROM_OPSERV_ONLY);
    helpserv_define_func("WRITE", cmd_write, HlOper, CMD_FROM_OPSERV_ONLY);
    helpserv_define_func("REGISTER", cmd_register, HlOper, CMD_FROM_OPSERV_ONLY);
    helpserv_define_func("MOVE", cmd_move, HlOper, CMD_NEED_BOT);
    helpserv_define_func("BOTS", cmd_bots, HlOper, CMD_FROM_OPSERV_ONLY|CMD_IGNORE_EVENT);

    helpserv_option_dict = dict_new();
    helpserv_define_option("PAGETARGET", opt_pagetarget_command);
    helpserv_define_option("ALERTPAGETARGET", opt_pagetarget_alert);
    helpserv_define_option("STATUSPAGETARGET", opt_pagetarget_status);
    helpserv_define_option("PAGE", opt_pagetype);
    helpserv_define_option("PAGETYPE", opt_pagetype);
    helpserv_define_option("ALERTPAGETYPE", opt_alert_page_type);
    helpserv_define_option("STATUSPAGETYPE", opt_status_page_type);
    helpserv_define_option("GREETING", opt_greeting);
    helpserv_define_option("OPENREQ", opt_openreq);
    helpserv_define_option("ASSIGNREQ", opt_assignreq);
    helpserv_define_option("CLOSEREQ", opt_closereq);
    helpserv_define_option("IDLEDELAY", opt_idle_delay);
    helpserv_define_option("WHINEDELAY", opt_whine_delay);
    helpserv_define_option("WHINEINTERVAL", opt_whine_interval);
    helpserv_define_option("EMPTYINTERVAL", opt_empty_interval);
    helpserv_define_option("STALEDELAY", opt_stale_delay);
    helpserv_define_option("REQPERSIST", opt_request_persistence);
    helpserv_define_option("HELPERPERSIST", opt_helper_persistence);
    helpserv_define_option("NOTIFICATION", opt_notification);

    helpserv_bots_dict = dict_new();
    dict_set_free_data(helpserv_bots_dict, helpserv_free_bot);
    
    helpserv_bots_bychan_dict = dict_new();
    dict_set_free_data(helpserv_bots_bychan_dict, helpserv_botlist_free);

    helpserv_reqs_bynick_dict = dict_new();
    dict_set_free_data(helpserv_reqs_bynick_dict, helpserv_reqlist_free);
    helpserv_reqs_byhand_dict = dict_new();
    dict_set_free_data(helpserv_reqs_byhand_dict, helpserv_reqlist_free);

    helpserv_users_byhand_dict = dict_new();
    dict_set_free_data(helpserv_users_byhand_dict, helpserv_userlist_free);

    alias_register_service("HelpServ", helpserv_command_search, helpserv_add_alias, helpserv_del_alias);

    helpserv_db_read();
    helpserv_helpfile_read();

    if (helpserv_conf.db_backup_frequency) {
        timeq_add(now + helpserv_conf.db_backup_frequency, helpserv_timed_write, NULL);
    }

    /* Make up for downtime... though this will only really affect the
     * time_per_week */
    if (helpserv_next_stats(last_stats_update) < now) {
        time_t statsrun=last_stats_update;

        while ((statsrun = helpserv_next_stats(statsrun)) < now) {
            helpserv_run_stats((void *)1);
        }
    }
    timeq_add(helpserv_next_stats(now), helpserv_run_stats, NULL);

    reg_join_func(handle_join);
    reg_part_func(handle_part); /* also deals with kick */
    reg_nick_change_func(handle_nickchange);
    reg_del_user_func(handle_quit);

    reg_auth_func(handle_nickserv_auth);
    reg_handle_rename_func(handle_nickserv_rename);
    reg_unreg_func(handle_nickserv_unreg);

    reg_exit_func(helpserv_db_cleanup);
    reg_db_write_func(helpserv_db_write);
}
