/* log.c - Diagnostic and error logging
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: log.c,v 1.39 2002/05/06 23:33:12 entrope Exp $
 */

#include "config.h"
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "common.h"
#include "log.h"
#include "helpfile.h"

extern int silent;

struct logData
{
    FILE		*file;
    char		*name;
    int                 severity_mask;
};

static struct logData logs[NUM_LOGS] =
{
    {NULL, NULL, 0},
    {NULL, "main.log", DEFAULT_SEVERITY_MASK},
#ifdef NDEBUG
    {NULL, "debug.log", 1 << LOG_ERROR | 1 << LOG_WARNING | 1 << LOG_OTHER},
#else
    {NULL, "debug.log", DEFAULT_SEVERITY_MASK},
#endif
    {NULL, "chanserv.log", DEFAULT_SEVERITY_MASK},
    {NULL, "opserv.log", DEFAULT_SEVERITY_MASK},
    {NULL, "nickserv.log", DEFAULT_SEVERITY_MASK},
    {NULL, "global.log", DEFAULT_SEVERITY_MASK},
    {NULL, "proxycheck.log", DEFAULT_SEVERITY_MASK},
    {NULL, "helpserv.log", DEFAULT_SEVERITY_MASK},
};

static struct logEntry logEntries[ENTRY_COUNT];
static unsigned int entries = 0;

#define MSG_INVALID_TYPE	"$b%s$b is an invalid log type."
#define MSG_INVALID_SEVERITY	"$b%s$b is an invalid severity level."

static void
cleanup_logs(void)
{
    unsigned int i;
    enum log_type lt;
    for (lt=1; lt<NUM_LOGS; lt++) {
	if (logs[lt].file) fclose(logs[lt].file);
    }
    for (i=0; i<ArrayLength(logEntries); i++) {
	free(logEntries[i].text);
	logEntries[i].text = 0;
	logEntries[i].size = 0;
    }
}

void
reopen_logs(void)
{
    enum log_type lt;
    for (lt=1; lt<NUM_LOGS;lt++) {
        if (logs[lt].file) {
            fclose(logs[lt].file);
        }
        logs[lt].file = fopen(logs[lt].name, "a");
    }
}

int
init_logs(void)
{
    enum log_type lt;
    int failed = 0;
    unsigned int i;
    for (lt=1; lt<NUM_LOGS; lt++) {
	logs[lt].file = fopen(logs[lt].name, "a");
	if (!logs[lt].file) failed = 1;
    }
    for(i = 0; i < ArrayLength(logEntries); i++) {
	logEntries[i].size = 80;
	logEntries[i].text = malloc(logEntries[i].size);
	logEntries[i].text[0] = 0;
    }
    reg_exit_func(cleanup_logs);
    return !failed;
}

void
log(enum log_type lt, enum log_severity ls, char *text, ...) {
    struct logEntry *entry;
    static char timestamp[24];
    va_list arg_list;
    const char *sev;
    unsigned int size;

    if (!(logs[lt].severity_mask & (1 << ls))) return;

    entry = logEntries + entries++;
    entry->type = lt;
    entry->severity = ls;

    if (entries == ArrayLength(logEntries)) {
	entries = 0;
    }

    entry->time = now;
    strftime(timestamp, sizeof(timestamp), "[%H:%M:%S %m/%d/%Y] ", localtime(&now));
    fputs(timestamp, logs[lt].file);
    switch (ls) {
    case LOG_INFO: sev = "(info) "; break;
    case LOG_WARNING: sev = "(warning) "; break;
    case LOG_ERROR: sev = "(error) "; break;
    case LOG_OTHER: default: sev = ""; break;
    }
    va_start(arg_list, text);
    if (logs[lt].file) {
        fputs(sev, logs[lt].file);
        size = vfprintf(logs[lt].file, text, arg_list) + 1;
        fflush(logs[lt].file);
    } else {
        size = vsnprintf(entry->text, 0, text, arg_list) + 1;
    }
    va_end(arg_list);
    if (!silent) {
        va_start(arg_list, text);
        vprintf(text, arg_list);
        va_end(arg_list);
    }
    if (size > entry->size) {
	entry->text = realloc(entry->text, size);
	entry->size = size;
    }
    va_start(arg_list, text);
    vsprintf(entry->text, text, arg_list);
    va_end(arg_list);
}

static enum log_type
find_log_type(const char *logname)
{
    if (!strcasecmp(logname, "main")) {
        return MAIN_LOG;
    } else if (!strcasecmp(logname, "debug")) {
        return DEBUG_LOG;
    } else if (!strcasecmp(logname, "chanserv")) {
        return CS_LOG;
    } else if (!strcasecmp(logname, "opserv")) {
        return OS_LOG;
    } else if (!strcasecmp(logname, "nickserv")) {
        return NS_LOG;
    } else if (!strcasecmp(logname, "global")) {
        return G_LOG;
    } else if (!strcasecmp(logname, "proxycheck")) {
        return PC_LOG;
    } else if (!strcasecmp(logname, "helpserv")) {
        return HS_LOG;
    } else {
        return NUM_LOGS;
    }
}

static enum log_severity
find_severity(const char *text, int len)
{
    if (!strncasecmp(text, "other", len)) {
        return LOG_OTHER;
    } else if (!strncasecmp(text, "info", len)) {
        return LOG_INFO;
    } else if (!strncasecmp(text, "warning", len)) {
        return LOG_WARNING;
    } else if (!strncasecmp(text, "error", len)) {
        return LOG_ERROR;
    } else {
        return LOG_NUM_SEVERITIES;
    }
}

struct logSearch*
log_discrim_create(struct userNode *service, struct userNode *user, unsigned int argc, unsigned char *argv[])
{
    unsigned int i;
    struct logSearch *discrim;

    /* Assume all criteria require arguments. */
    if((argc - 1) % 2)
    {
	send_message(user, service, MSG_MISSING_PARAMS, argv[0]);
	return NULL;
    }

    discrim = malloc(sizeof(struct logSearch));
    memset(discrim, 0, sizeof(*discrim));
    discrim->limit = 25;

    for (i=1; i<argc-1; i++) {
	if (!strcasecmp(argv[i], "severity")) {
            i++;
            discrim->severity = find_severity(argv[i], strlen(argv[i]));
            if (discrim->severity == LOG_NUM_SEVERITIES) {
		send_message(user, service, MSG_INVALID_SEVERITY, argv[i]);
		goto fail;
	    }
	} else if (strcasecmp(argv[i], "type") == 0) {
            discrim->type = find_log_type(argv[++i]);
            if (discrim->type == NUM_LOGS) {
		send_message(user, service, MSG_INVALID_TYPE, argv[i]);
		goto fail;
	    }
	} else if (strcasecmp(argv[i], "text") == 0) {
	    discrim->glob = argv[++i];
	} else if (strcasecmp(argv[i], "limit") == 0) {
	    discrim->limit = strtoul(argv[++i], NULL, 10);
	} else {
	    send_message(user, service, MSG_INVALID_CRITERIA, argv[i]);
	    goto fail;
	}
    }

    return discrim;
  fail:
    free(discrim);
    return NULL;
}

static int
entry_match(struct logSearch *discrim, struct logEntry *entry)
{
    if ((discrim->type && discrim->type != entry->type)
	|| (discrim->severity && discrim->severity != entry->severity)
        || (discrim->glob && !match_ircglob(entry->text, discrim->glob))) {
	return 0;
    }
    return 1;
}

unsigned int
log_entry_search(struct logSearch *discrim, entry_search_func esf, void *data)
{
    unsigned int matched = 0, start = entries, m;
    struct logEntry *entry;

    for (m = 0; m < ArrayLength(logEntries); m++) {
        entry = logEntries + start - m;
        if (m > start) {
            entry = entry + ArrayLength(logEntries);
        }

        if (!entry->text[0]) continue;

	if (entry_match(discrim, entry)) {
	    esf(entry, data);
	    if (++matched >= discrim->limit) break;
	}
    }

    return matched;
}

int
log_modify_logging(const char *logname, const char *mods)
{
    enum log_type target;
    enum log_severity bit;
    int mod_type, new_mask;
    size_t mod_len;

    target = find_log_type(logname);
    if (target == NUM_LOGS) return 0;

    switch (mods[0]) {
    case '+':
        new_mask = logs[target].severity_mask;
        mod_type = 1;
        mods++;
        break;
    case '-':
        new_mask = logs[target].severity_mask;
        mod_type = 0;
        mods++;
        break;
    case '>':
        if (mods[1] == '=') {
            mod_type = 1;
            mods += 2;
        } else {
            mod_type = 0;
            mods++;
        }
        bit = find_severity(mods, strlen(mods));
        if (bit == LOG_NUM_SEVERITIES) return 0;
        if (!mod_type) bit++;
        logs[target].severity_mask = 0;
        while (bit < LOG_NUM_SEVERITIES) {
            logs[target].severity_mask |= (1 << bit++);
        }
        return 1;
    case '<':
        if (mods[1] == '=') {
            mod_type = 1;
            mods += 2;
        } else {
            mod_type = 0;
            mods++;
        }
        bit = find_severity(mods, strlen(mods));
        if (bit == LOG_NUM_SEVERITIES) return 0;
        if (mod_type) bit++;
        logs[target].severity_mask = (1 << bit) - 1;
        return 1;
    default:
        new_mask = 0;
        mod_type = 1;
    }

    while (*mods) {
        mod_len = strcspn(mods, "+-");
        bit = find_severity(mods, mod_len);
        if (bit == LOG_NUM_SEVERITIES) return 0;
        if (mod_type) {
            logs[target].severity_mask |= 1 << bit;
        } else {
            logs[target].severity_mask &= ~(1 << bit);
        }
        mods += mod_len;
        switch (mods[0]) {
        case '+': mod_type = 1; mods++; break;
        case '-': mod_type = 0; mods++; break;
        }
    }
    return 1;
}

int
log_get_mask(const char *logname)
{
    enum log_type target = find_log_type(logname);
    return (target == NUM_LOGS) ? 0 : logs[target].severity_mask;
}
