/* main.c - srvx
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: main.c,v 1.123 2002/05/19 14:14:33 entrope Exp $
 */

#define PID_FILE "srvx.pid"

#include "config.h"
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
#ifdef HAVE_SYS_RESOURCE_H
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/resource.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifndef SIGCHLD
#define SIGCHLD SIGCLD
#endif

#include "alias.h"
#include "common.h"
#include "conf.h"
#include "dict.h"
#include "gline.h"
#include "hash.h"
#include "ioset.h"
#include "log.h"
#include "parse.h"
#include "policer.h"
#include "recdb.h"
#include "sendmail.h"
#include "timeq.h"

#include "chanserv.h"
#include "global.h"
#include "nickserv.h"
#include "opserv.h"
#include "sockcheck.h"
#include "helpserv.h"

extern FILE *replay_file;

time_t boot_time = 0, burst_begin = 0, now;
unsigned long burst_length = 0;

int ping_freq = 120, ping_timeout = 30;
int silent, quit_services, max_cycles;

char *services_config = "srvx.conf";

char **services_argv;
int services_argc;

struct cManagerNode cManager;

struct policer_params *oper_policer_params, *luser_policer_params, *god_policer_params;

void uplink_select(char *name);

static int
uplink_insert(const char *key, void *data, void *extra)
{
    struct uplinkNode *uplink = malloc(sizeof(struct uplinkNode));
    struct record_data *rd = data;
    int enabled = 1;
    char *str;
    struct sockaddr_in *sin;
    unsigned long addr;
    (void)extra;

    if(!uplink)
    {
	return 0;
    }

    uplink->name = (char *)key;
    uplink->host = database_get_data(rd->d.object, "address", RECDB_QSTRING);

    str = database_get_data(rd->d.object, "port", RECDB_QSTRING);
    uplink->port = str ? atoi(str) : 6667;
    uplink->password = database_get_data(rd->d.object, "password", RECDB_QSTRING);
    uplink->their_password = database_get_data(rd->d.object, "uplink_password", RECDB_QSTRING);

    str = database_get_data(rd->d.object, "enabled", RECDB_QSTRING);
    if(str)
    {
	enabled = atoi(str) ? 1 : 0;
    }

    cManager.enabled += enabled;

    str = database_get_data(rd->d.object, "max_tries", RECDB_QSTRING);
    uplink->max_tries = str ? atoi(str) : 3;
    uplink->flags = enabled ? 0 : UPLINK_UNAVAILABLE;
    uplink->state = DISCONNECTED;
    uplink->tries = 0;

    str = database_get_data(rd->d.object, "bind_address", RECDB_QSTRING);
    uplink->bind_addr_len = sizeof(*sin);
    if (str && getipbyname(str, &addr)) 
    {
	sin = malloc(uplink->bind_addr_len);
        sin->sin_port = 0;
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = addr;
#ifdef HAVE_SIN_LEN
	sin->sin_len = 0;
#endif
	memset(sin->sin_zero, 0, sizeof(sin->sin_zero));
	uplink->bind_addr = sin;
    } 
    else 
    {
	uplink->bind_addr = NULL;
	uplink->bind_addr_len = 0;
    }

    uplink->next = cManager.uplinks;
    uplink->prev = NULL;

    if(cManager.uplinks)
    {
	cManager.uplinks->prev = uplink;
    }

    cManager.uplinks = uplink;

    /* If the configuration is being reloaded, set the current uplink
       to the reloaded equivalent, if possible.

       XXX: May want to include the bind_address in the equivalence test. */
    if(cManager.uplink && enabled && !strcasecmp(uplink->host, cManager.uplink->host) && uplink->port == cManager.uplink->port)
    {
	uplink->state = cManager.uplink->state;
	uplink->tries = cManager.uplink->tries;
	cManager.uplink = uplink;
    }

    return 0;
}

void uplink_compile(void)
{
    const char *cycles;
    dict_t conf_node;
    struct uplinkNode *oldUplinks = NULL, *oldUplink = NULL;

    /* Save the old uplinks, we'll remove them later. */
    oldUplink = cManager.uplink;
    oldUplinks = cManager.uplinks;

    cycles = conf_get_data("server/max_cycles", RECDB_QSTRING);
    max_cycles = cycles ? atoi(cycles) : 30;
    if(!(conf_node = conf_get_data("uplinks", RECDB_OBJECT)))
    {
        log(MAIN_LOG, LOG_ERROR, "No uplinks configured; giving up.\n");
	exit(1);
    }

    cManager.enabled = 0;
    dict_foreach(conf_node, uplink_insert, NULL);

    /* Remove the old uplinks, if any. It doesn't matter if oldUplink (below)
       is a reference to one of these, because it won't get dereferenced. */
    if(oldUplinks)
    {
	struct uplinkNode *uplink, *next;

	oldUplinks->prev->next = NULL;

	for(uplink = oldUplinks; uplink; uplink = next)
	{
	    next = uplink->next;
	    free(uplink);
	}
    }

    /* If the uplink hasn't changed, it's either NULL or pointing at
       an uplink that was just deleted, select a new one. */
    if(cManager.uplink == oldUplink)
    {
	if(oldUplink)
	{
	    irc_squit(self, "Uplinks updated; selecting new uplink.");
	}

	cManager.uplink = NULL;
	uplink_select(NULL);
    }
}

int uplink_valid(char *name)
{
    struct uplinkNode *uplink;

    if(!cManager.enabled || !cManager.uplinks)
    {
	return 0;
    }

    for(uplink = cManager.uplinks; uplink; uplink = uplink->next)
    {
	if(!strcmp(uplink->name, name))
	{
	    return 1;
	}
    }

    return 0;
}

void uplink_select(char *name)
{
    struct uplinkNode *start, *uplink, *next;
    int stop;

    if(!cManager.enabled || !cManager.uplinks)
    {
	log(MAIN_LOG, LOG_ERROR, "No uplinks enabled; giving up.\n");
	exit(1);
    }

    if(!cManager.uplink)
    {
	start = cManager.uplinks;
    }
    else
    {
	start = cManager.uplink->next;
	if(!start)
	{
	    start = cManager.uplinks;
	}
    }

    stop = 0;
    for(uplink = start; uplink; uplink = next)
    {
	next = uplink->next ? uplink->next : cManager.uplinks;

	if(stop)
	{
	    uplink = NULL;
	    break;
	}

	/* We've wrapped around the list. */
	if(next == start)
	{
	    sleep((cManager.cycles >> 1) * 5);
	    cManager.cycles++;

	    if(max_cycles && (cManager.cycles > max_cycles))
	    {
		log(MAIN_LOG, LOG_ERROR, "Maximum uplink list cycles exceeded; giving up.\n");
		exit(1);
	    }

	    /* Give the uplink currently in 'uplink' consideration,
	       and if not selected, break on the next iteration. */
	    stop = 1;
	}

	/* Skip bad uplinks. */
	if(uplink->flags & UPLINK_UNAVAILABLE)
	{
	    continue;
	}

	if(name && strcasecmp(uplink->name, name))
	{
	    /* If we were told to connect to a specific uplink, don't stop
	       until we find it.
	    */
	    continue;
	}

	/* It would be possible to track uplink health through a variety
	   of statistics and only break on the best uplink. For now, break
	   on the first available one.
	*/

	break;
    }

    if(!uplink)
    {
	/* We are shit outta luck if every single uplink has been passed
	   over. Use the current uplink if possible. */
	if(!cManager.uplink || cManager.uplink->flags & UPLINK_UNAVAILABLE)
	{
	    log(MAIN_LOG, LOG_ERROR, "All available uplinks exhausted; giving up.\n");
	    exit(1);
	}

	return;
    }

    cManager.uplink = uplink;
}

int uplink_connect(void)
{
    struct uplinkNode *uplink = cManager.uplink;

    if(uplink->state != DISCONNECTED)
    {
	return 0;
    }

    if(uplink->flags & UPLINK_UNAVAILABLE)
    {
	uplink_select(NULL);
	uplink = cManager.uplink;
    }

    if(uplink->tries)
    {
	/* This delay could scale with the number of tries. */
	sleep(2);
    }

    if(!create_socket_client(uplink))
    {
	if(uplink->max_tries && (uplink->tries > uplink->max_tries))
	{
	    /* This is a bad uplink, move on. */
	    uplink->flags |= UPLINK_UNAVAILABLE;
	    uplink_select(NULL);
	}

	return 0;
    }
    else
    {
	uplink->state = AUTHENTICATING;
	irc_introduce(uplink->password);
    }

    return 1;
}

void
received_ping(void)
{
    /* This function is called when a ping is received. Take it as
       a sign of link health and reset the connection manager
       information. */

    cManager.cycles = 0;
}

void sigaction_writedb(int x)
{
#ifndef HAVE_STRSIGNAL
    log(MAIN_LOG, LOG_INFO, "Signal %d -- writing databases.\n", x);
#else
    log(MAIN_LOG, LOG_INFO, "%s -- writing databases.\n", strsignal(x));
#endif
    write_databases();
}

void sigaction_exit(int x)
{
#ifndef HAVE_STRSIGNAL
    log(MAIN_LOG, LOG_INFO, "Signal %d -- exiting.\n", x);
#else
    log(MAIN_LOG, LOG_INFO, "%s -- exiting.\n", strsignal(x));
#endif
    irc_squit(self, "Exiting on signal from console.");
    quit_services = 1;
}

void sigaction_wait(int x)
{
    int code;
    (void)x;
    wait(&code);
}

void sigaction_rehash(int x)
{
#ifndef HAVE_STRSIGNAL
    log(MAIN_LOG, LOG_INFO, "Signal %d -- rehashing.\n", x);
#else
    log(MAIN_LOG, LOG_INFO, "%s -- rehashing.\n", strsignal(x));
#endif
    conf_read(services_config);
    reopen_logs();
}

static exit_func_t *ef_list;
static unsigned int ef_size = 0, ef_used = 0;

void reg_exit_func(exit_func_t handler)
{
    if (ef_used == ef_size) {
	if (ef_size) {
	    ef_size <<= 1;
	    ef_list = realloc(ef_list, ef_size*sizeof(exit_func_t));
	} else {
	    ef_size = 8;
	    ef_list = malloc(ef_size*sizeof(exit_func_t));
	}
    }
    ef_list[ef_used++] = handler;
}

void call_exit_funcs(void)
{
    unsigned int n = ef_used;

    /* Call them in reverse order because we initialize logs, then
     * nickserv, then chanserv, etc., and they register their exit
     * funcs in that order, and there are some dependencies (for
     * example, ChanServ requires NickServ to not have cleaned up).
     */

    while (n > 0) {
	ef_list[--n]();
    }
    if (ef_list) free(ef_list);
    ef_used = ef_size = 0;
}

static db_write_func_t *dw_list;
static unsigned int dw_size = 0, dw_used = 0;

void reg_db_write_func(db_write_func_t handler)
{
    if (dw_used == dw_size) {
	if (dw_size) {
	    dw_size <<= 1;
	    dw_list = realloc(dw_list, dw_size*sizeof(db_write_func_t));
	} else {
	    dw_size = 8;
	    dw_list = malloc(dw_size*sizeof(db_write_func_t));
	}
    }
    dw_list[dw_used++] = handler;
}

void write_databases(void)
{
    unsigned int n = dw_used;

    /* Order shouldn't matter, and this is a cut-n-paste hack job anyway. */
    while (n > 0) {
	dw_list[--n]();
    }
}

static void
conf_globals(void)
{
    const char *info;
    dict_t dict;

    info = conf_get_data("services/global/nick", RECDB_QSTRING);
    if (info && (info[0] != '.')) init_global(info);

    info = conf_get_data("services/nickserv/nick", RECDB_QSTRING);
    if (info && (info[0] != '.')) init_nickserv(info);

    info = conf_get_data("services/chanserv/nick", RECDB_QSTRING);
    if (info && (info[0] != '.')) init_chanserv(info);

    info = conf_get_data("services/helpserv/enable", RECDB_QSTRING);
    if (!info || !disabled_string(info)) init_helpserv();

    god_policer_params = policer_params_new();
    if ((dict = conf_get_data("policers/commands-god", RECDB_OBJECT))) {
        dict_foreach(dict, set_policer_param, god_policer_params);
    } else {
        policer_params_set(god_policer_params, "size", "30");
        policer_params_set(god_policer_params, "drain-rate", "1");
    }
    oper_policer_params = policer_params_new();
    if ((dict = conf_get_data("policers/commands-oper", RECDB_OBJECT))) {
        dict_foreach(dict, set_policer_param, oper_policer_params);
    } else {
        policer_params_set(oper_policer_params, "size", "10");
        policer_params_set(oper_policer_params, "drain-rate", "1");
    }
    luser_policer_params = policer_params_new();
    if ((dict = conf_get_data("policers/commands-luser", RECDB_OBJECT))) {
        dict_foreach(dict, set_policer_param, luser_policer_params);
    } else {
        policer_params_set(luser_policer_params, "size", "3");
        policer_params_set(luser_policer_params, "drain-rate", "0.10");
    }

    info = conf_get_data("services/opserv/nick", RECDB_QSTRING);
    if (info) init_opserv(info);
}

#ifdef HAVE_SYS_RESOURCE_H

static int
set_item_rlimit(const char *name, void *data, void *extra)
{
    int rsrc, found;
    struct record_data *rd = data;
    struct rlimit rlim;
    const char *str;

    rsrc = (int)dict_find(extra, name, &found);
    if (!found) {
        log(MAIN_LOG, LOG_ERROR, "Invalid rlimit \"%s\" in rlimits section.\n", name);
        return 0;
    }
    if (!(str = GET_RECORD_QSTRING(rd))) {
        log(MAIN_LOG, LOG_ERROR, "Missing or invalid parameter type for rlimit \"%s\".\n", name);
        return 0;
    }
    if (getrlimit(rsrc, &rlim) < 0) {
        log(MAIN_LOG, LOG_ERROR, "Couldn't get rlimit \"%s\": errno %d: %s\n", name, errno, strerror(errno));
        return 0;
    }
    rlim.rlim_cur = ParseVolume(str);
    if (setrlimit(rsrc, &rlim) < 0) {
        log(MAIN_LOG, LOG_ERROR, "Couldn't set rlimit \"%s\": errno %d: %s\n", name, errno, strerror(errno));
    }
    return 0;
}

static void
conf_rlimits(void)
{
    dict_t dict, values;

    values = dict_new();
    dict_insert(values, "data", (void*)RLIMIT_DATA);
    dict_insert(values, "stack", (void*)RLIMIT_STACK);
#ifdef RLIMIT_VMEM
    dict_insert(values, "vmem", (void*)RLIMIT_VMEM);
#else
#ifdef RLIMIT_AS
    dict_insert(values, "vmem", (void*)RLIMIT_AS);
#endif
#endif
    if ((dict = conf_get_data("rlimits", RECDB_OBJECT))) {
        dict_foreach(dict, set_item_rlimit, values);
    }
    dict_delete(values);
}

#else

static void
conf_rlimits(void)
{
}

#endif

void main_shutdown(void)
{
    struct uplinkNode *ul, *ul_next;
    ioset_cleanup();
    for (ul = cManager.uplinks; ul; ul = ul_next) {
        ul_next = ul->next;
        if (ul->bind_addr) free(ul->bind_addr);
        free(ul);
    }
    tools_cleanup();
    conf_close();
    unlink(PID_FILE);
    if (dw_list) free(dw_list);
    policer_params_delete(god_policer_params);
    policer_params_delete(oper_policer_params);
    policer_params_delete(luser_policer_params);
    if (replay_file) fclose(replay_file);
}

void usage(char *self) {
    /* We can assume we have getopt_long(). */
    printf("Usage: %s [-c config] [-r log] [-f] [-V] [-v|-h]\n\
-c, --config			selects a different configuration file.\n\
-f, --foreground		run srvx in the foreground.\n\
-h, --help			prints this usage message.\n\
-k, --check			checks the configuration file's syntax.\n\
-r, --replay			replay a log file (debug)\n\
-v, --version			prints this program's version.\n\
-V, --verbose			send log messages to the console.\n", self);
}

void version() {
    printf("    --------------------------------------------------\n\
    - srvx "VERSION" ("CODENAME"), Built: " __DATE__ ", " __TIME__".\n\
    - Copyright (C) 2000 - 2002, srvx Development Team\n\
    --------------------------------------------------\n");
}

void license() {
    printf("\n\
    This program is free software; you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation; either version 2 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program; if not, write to the Free Software\n\
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n\n");
}

#if WITH_MALLOC_BOEHM_GC
void
gc_warn_proc(char *msg, GC_word arg)
{
    log(MAIN_LOG, LOG_ERROR, "GC(%p): %s\n", arg, msg);
}
#endif

int main(int argc, char *argv[])
{
    int tmp, daemon;
    pid_t pid = 0;
    FILE *file_out;
    struct sigaction sv;

    daemon = 1;
    silent = 1;

    /* set up some signal handlers */
    memset(&sv, 0, sizeof(sv));
    sigemptyset(&sv.sa_mask);
    sigaddset(&sv.sa_mask, SIGPIPE);
    sv.sa_handler = sigaction_rehash;
    sigaction(SIGHUP, &sv, NULL);
    sv.sa_handler = sigaction_writedb;
    sigaction(SIGINT, &sv, NULL);
    sv.sa_handler = sigaction_exit;
    sigaction(SIGQUIT, &sv, NULL);
    sv.sa_handler = sigaction_wait;
    sigaction(SIGCHLD, &sv, NULL);

    if (argc > 1) { /* parse command line, if any */
	int c;
	struct option options[] =
	{
	    {"config", 1, 0, 'c'},
	    {"foreground", 0, 0, 'f'},
	    {"help", 0, 0, 'h'},
	    {"check", 0, 0, 'k'},
            {"replay", 1, 0, 'r'},
	    {"version", 0, 0, 'v'},
	    {"verbose", 0, 0, 'V'},
	    {0, 0, 0, 0}
	};

	while ((c = getopt_long(argc, argv, "c:fhkr:vV", options, NULL)) != -1) {
	    switch(c) {
	    case 'c':
		services_config = optarg;
		break;
	    case 'k':
		if(conf_read(services_config))
		{
		    printf("%s appears to be a valid configuration file.\n", services_config);
		}
		else
		{
		    printf("%s is an invalid configuration file.\n", services_config);
		}
		exit(0);
            case 'r':
                replay_file = fopen(optarg, "r");
                if (!replay_file) {
                    fprintf(stderr, "Could not open %s for reading: %s (%d)\n",
                            optarg, strerror(errno), errno);
                    exit(0);
                }
                break;
	    case 'f':
		daemon = 0;
		break;
	    case 'v':
		version();
		license();
		exit(0);
	    case 'V':
		silent = 0;
		break;
	    case 'h':
	    default:
		usage(argv[0]);
		exit(0);
	    }
	}
    }

    if(daemon)
    {
	silent = 1;
    }

    version();

#ifdef __CYGWIN__
    if(daemon)
    {
	fprintf(stderr, "Daemon mode not supported under Cygwin.\n");
	daemon = 0;
    }
#endif
#if WITH_MALLOC_BOEHM_GC
    GC_set_warn_proc(gc_warn_proc);
#endif

    if (replay_file) {
        /* We read a line here to "prime" the replay file parser, but
         * mostly to get the right value of "now" for when we do the
         * irc_introduce. */
        replay_read_line();
        boot_time = now;
    } else {
        boot_time = now = time(NULL);
    }
    init_logs();

    log(MAIN_LOG, LOG_INFO, "Initializing daemon...\n");
    if (!conf_read(services_config)) {
	log(MAIN_LOG, LOG_ERROR, "Unable to read %s.\n", services_config);
	exit(0);
    }

    uplink_compile();
    conf_register_reload(uplink_compile);

    if (daemon) {
	/* Attempt to fork into the background if daemon mode is on. */
	pid = fork();
	if(pid == -1)
	    log(MAIN_LOG, LOG_ERROR, "Unable to fork: %s\n", strerror(errno));
	else if(pid > 0)
	{
	    int temp_silent = silent;
	    silent = 0;
	    log(MAIN_LOG, LOG_INFO, "Forking into the background (pid: %i)...\n\n", pid);
	    silent = temp_silent;

	    exit(0);
	}
	setsid();
    }

    if ((file_out = fopen(PID_FILE, "w")) == NULL) {
	/* Create the main process' pid file */
	log(MAIN_LOG, LOG_ERROR, "Unable to create PID file: %s\n", strerror(errno));
    } else {
	fprintf(file_out, "%i\n", (int)getpid());
	fclose(file_out);
    }

    services_argc = argc;
    services_argv = argv;

    atexit(call_exit_funcs);
    reg_exit_func(main_shutdown);

    tmp = conf_server_ping_freq();
    if (tmp != -1) ping_freq = tmp;
    tmp = conf_server_ping_timeout();
    if (tmp != -1) ping_timeout = tmp;
    timeq_init();
    sendmail_init();
    gline_init();
    sockcheck_init();
    init_structs();
    init_parse();
    alias_init();
    conf_globals();
    conf_rlimits();
    read_aliases();
    if (replay_file) log(MAIN_LOG, LOG_INFO, "Beginning replay...\n");

    /* The first exit func to be called *should* be write_databases(). */
    reg_exit_func(write_databases);

    if (!replay_file) now = time(NULL);
    srand(now);
    if (replay_file) {
        replay_event_loop();
    } else {
        ioset_run();
    }
    return 0;
}
